import uuid
from functools import reduce

import altair as alt
from IPython.display import HTML, display

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.probes.utils import CARE_SITE_LEVEL_NAMES
from edsteva.viz.utils import filter_predictor, round_it, save_html, scale_it


def estimates_dashboard(
    probe: BaseProbe,
    fitted_model: BaseModel,
    care_site_level: str = CARE_SITE_LEVEL_NAMES["Hospital"],
    save_path: str = None,
    labelFontSize: float = 12,
    titleFontSize: float = 13,
):
    r"""Displays an interactive chart with:

    - On the top, the aggregated normalized completeness predictor $\frac{c(\Delta t)}{c_0}$ over normalized time $\Delta t = t - t_0$. It represents the overall deviation from the Model.
    - On the bottom, interactive filters including all the columns in the [Probe][probe] (such as time, care site, number of visits...etc.) and all the estimates (coefficients and metrics) in the [Model][model].

    Is is possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$
    fitted_model : BaseModel
        Model fitted to the probe
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    save_path : str, optional
        Folder path where to save the chart in HTML format.

        **EXAMPLE**: `"my_folder/my_file.html"`
    labelFontSize: float, optional
        The font size of the labels (axis and legend).
    titleFontSize: float, optional
        The font size of the titles.
    """
    alt.data_transformers.disable_max_rows()

    predictor = probe.predictor.copy()
    estimates = fitted_model.estimates.copy()
    predictor = predictor.merge(estimates, on=probe._index)

    def month_diff(x, y):
        end = x.dt.to_period("M").view(dtype="int64")
        start = y.dt.to_period("M").view(dtype="int64")
        return end - start

    predictor["date"] = predictor["date"].astype("datetime64[ns]")
    predictor["t_0"] = predictor["t_0"].astype("datetime64[ns]")
    predictor["normalized_date"] = month_diff(predictor["date"], predictor["t_0"])
    predictor["t_0"] = predictor["t_0"].astype(str)
    predictor["normalized_date"] = predictor["normalized_date"].astype(int)

    predictor["normalized_c"] = predictor["c"].mask(
        (predictor["normalized_date"] >= 0) & (predictor["c_0"] == 0), 1
    )
    predictor["normalized_c"] = predictor["normalized_c"].mask(
        (predictor["normalized_date"] >= 0) & (predictor["c_0"] > 0),
        predictor["c"] / predictor["c_0"],
    )

    predictor["legend_error_band"] = "Standard deviation"
    predictor["legend_model"] = type(fitted_model).__name__

    predictor["model"] = predictor["c_0"].where(predictor["normalized_date"] < 0, 1)
    predictor["model"] = predictor["model"].where(predictor["normalized_date"] >= 0, 0)

    predictor = filter_predictor(
        predictor=predictor,
        care_site_level=care_site_level,
    )

    # RectangleModel
    if fitted_model.name == "RectangleFunction":
        predictor = predictor[predictor.date < predictor.t_1]

    index = list(set(probe._index).difference(["care_site_level", "care_site_id"]))
    time = "date"
    _predictor = "c"

    c_0_min_slider = alt.binding_range(
        min=0,
        max=round_it(predictor.c_0.max(), 2),
        step=scale_it(predictor.c_0.max()) / 100,
        name="c₀ min: ",
    )
    c_0_min_selection = alt.selection_single(
        name="c_0_min",
        fields=["c_0_min"],
        bind=c_0_min_slider,
        init={"c_0_min": 0},
    )
    t_0_slider = alt.binding(
        input="t_0",
        name="t₀ max: ",
    )
    t_0_selection = alt.selection_single(
        name="t_0",
        fields=["t_0"],
        bind=t_0_slider,
        init={"t_0": predictor.t_0.astype(str).max()},
    )
    if fitted_model.name == "RectangleFunction":
        t_1_slider = alt.binding(
            input="t_1",
            name="t₁ min: ",
        )
        t_1_selection = alt.selection_single(
            name="t_1",
            fields=["t_1"],
            bind=t_1_slider,
            init={"t_1": predictor.t_1.astype(str).min()},
        )

    error_max_slider = alt.binding_range(
        min=0,
        max=round_it(predictor.error.max(), 2),
        step=scale_it(predictor.error.max()) / 100,
        name="error max: ",
    )
    error_max_selection = alt.selection_single(
        name="error_max",
        fields=["error_max"],
        bind=error_max_slider,
        init={"error_max": round_it(predictor.error.max(), 2)},
    )

    care_site_selection = alt.selection_multi(fields=["care_site_short_name"])
    care_site_color = alt.condition(
        care_site_selection,
        alt.Color(
            "care_site_short_name:N",
            legend=None,
            sort={"field": "c_0", "op": "min", "order": "descending"},
        ),
        alt.value("lightgray"),
    )

    base_chart = alt.Chart(predictor).transform_joinaggregate(
        mean_c_0="mean(c_0)",
        mean_error="mean(error)",
        groupby=["care_site_short_name"] + index,
    )

    time_selection = alt.selection_interval(encodings=["x"])
    time_line = (
        base_chart.mark_line()
        .encode(
            x=alt.X(
                "normalized_date:Q",
                title="Δt = (t - t₀) months",
                scale=alt.Scale(nice=False),
            ),
            y=alt.Y(
                "mean(c):Q",
                title="c(Δt)",
            ),
        )
        .add_selection(time_selection)
    ).properties(width=800, height=50)

    predictor_hist = (
        base_chart.mark_bar()
        .encode(
            y=alt.Y(
                "care_site_short_name:N",
                title="Care site short name",
                sort="-x",
            ),
            color=care_site_color,
        )
        .add_selection(care_site_selection)
        .transform_filter(alt.datum.mean_c_0 >= c_0_min_selection.c_0_min)
        .transform_filter(alt.datum.mean_error <= error_max_selection.error_max)
        .transform_filter(alt.datum.t_0 <= t_0_selection.t_0)
    ).properties(width=300)

    # RectangleModel
    if fitted_model.name == "RectangleFunction":
        predictor_hist = predictor_hist.transform_filter(
            alt.datum.t_1 >= t_1_selection.t_1
        )

    c_0_hist = predictor_hist.encode(
        x=alt.X(
            "min(mean_c_0):Q",
            title="Min(c₀)",
        ),
        tooltip=alt.Tooltip("min(mean_c_0):Q", format=".2"),
    )
    error_hist = predictor_hist.encode(
        x=alt.X(
            "max(mean_error):Q",
            title="Max(error)",
        ),
        tooltip=alt.Tooltip("max(mean_error):Q", format=".2"),
    )

    index_variables_hists = []
    index_variables_selections = []
    for index_variable in index:
        index_variable_selection = alt.selection_multi(fields=[index_variable])
        index_variables_selections.append(index_variable_selection)

        index_variable_color = alt.condition(
            index_variable_selection,
            alt.Color(
                "{}:N".format(index_variable),
                legend=None,
                sort={"field": "c_0", "op": "min", "order": "descending"},
            ),
            alt.value("lightgray"),
        )
        index_variable_hist = (
            base_chart.mark_bar()
            .encode(
                y=alt.Y(
                    "mean(c):Q",
                    title="c",
                ),
                x=alt.X(
                    "{}:N".format(index_variable),
                    title=index_variable,
                    sort="-y",
                ),
                tooltip=alt.Tooltip("mean(c):Q", format=".2"),
                color=index_variable_color,
            )
            .add_selection(index_variable_selection)
            .transform_filter(care_site_selection)
            .transform_filter(alt.datum.mean_c_0 >= c_0_min_selection.c_0_min)
            .transform_filter(alt.datum.t_0 <= t_0_selection.t_0)
            .transform_filter(alt.datum.mean_error <= error_max_selection.error_max)
        ).properties(title="Average completeness per {}".format(index_variable))

        # RectangleModel
        if fitted_model.name == "RectangleFunction":
            index_variable_hist = index_variable_hist.transform_filter(
                alt.datum.t_1 >= t_1_selection.t_1
            )

        index_variables_hists.append(index_variable_hist)

    extra_predictors = predictor.columns.difference(
        index
        + [time]
        + [_predictor]
        + [
            "care_site_short_name",
            "care_site_id",
            "model",
            "legend_error_band",
            "legend_model",
            "normalized_date",
            "normalized_c",
            "c_0",
            "t_0",
            "t_1",
            "error",
        ]
    )
    extra_predictors_hists = []

    for extra_predictor in extra_predictors:

        extra_predictor_hist = predictor_hist.encode(
            x=alt.X(
                "sum({}):Q".format(extra_predictor),
                title="{}".format(extra_predictor),
                axis=alt.Axis(format="s"),
            ),
            tooltip=alt.Tooltip("sum({}):Q".format(extra_predictor), format=","),
        ).properties(title="Total {} per care site".format(extra_predictor))
        extra_predictors_hists.append(extra_predictor_hist)

    index.append("care_site_short_name")
    index_selection = alt.selection_single(
        fields=["index"],
        bind=alt.binding_radio(name="Plot average completeness per: ", options=index),
        init={"index": "stay_type"},
    )

    probe_line = (
        base_chart.transform_fold(index, as_=["index", "value"])
        .encode(
            x=alt.X(
                "normalized_date:Q",
                title="Δt = (t - t₀) months",
                scale=alt.Scale(nice=False),
            ),
            color=alt.Color(
                "value:N",
                sort={"field": "c_0", "op": "min", "order": "descending"},
                title=None,
            ),
        )
        .transform_filter(alt.datum.mean_c_0 >= c_0_min_selection.c_0_min)
        .transform_filter(alt.datum.mean_error <= error_max_selection.error_max)
        .transform_filter(alt.datum.t_0 <= t_0_selection.t_0)
        .transform_filter(care_site_selection)
        .transform_filter(index_selection)
    ).properties(width=900, height=300)
    # RectangleModel
    if fitted_model.name == "RectangleFunction":
        probe_line = probe_line.transform_filter(alt.datum.t_1 >= t_1_selection.t_1)

    mean_line = probe_line.mark_line().encode(
        y=alt.Y(
            "mean(normalized_c):Q",
        )
    )
    error_line = probe_line.mark_errorband(extent="stdev").encode(
        y=alt.Y(
            "normalized_c:Q",
            title="c(Δt) / c₀",
        ),
        stroke=alt.Stroke(
            "legend_error_band",
            title="Error band",
            legend=alt.Legend(symbolType="square", orient="top"),
        ),
    )
    model_line = (
        alt.Chart(predictor)
        .mark_line(color="black", interpolate="step-after", strokeDash=[5, 5])
        .encode(
            x=alt.X(
                "normalized_date:Q",
                scale=alt.Scale(nice=False),
            ),
            y="model:Q",
            strokeWidth=alt.StrokeWidth(
                "legend_model",
                title="Model line",
                legend=alt.Legend(orient="top", symbolDash=[2, 2]),
            ),
        )
    )

    top_chart = mean_line + error_line + model_line

    top_chart = top_chart.add_selection(index_selection).transform_filter(
        time_selection
    )

    for index_variable_selection in index_variables_selections:
        top_chart = top_chart.transform_filter(index_variable_selection)
        c_0_hist = c_0_hist.transform_filter(index_variable_selection)
        error_hist = error_hist.transform_filter(index_variable_selection)
        for idx in range(len(extra_predictors_hists)):
            extra_predictors_hists[idx] = extra_predictors_hists[idx].transform_filter(
                index_variable_selection
            )
        for idx in range(len(index_variables_hists)):
            if idx != index_variables_selections.index(index_variable_selection):
                index_variables_hists[idx] = index_variables_hists[
                    idx
                ].transform_filter(index_variable_selection)

    index_variables_hists = reduce(
        lambda index_variable_hist_1, index_variable_hist_2: index_variable_hist_1
        & index_variable_hist_2,
        index_variables_hists,
    )
    extra_predictors_hists = reduce(
        lambda extra_predictor_hist_1, extra_predictor_hist_2: extra_predictor_hist_1
        & extra_predictor_hist_2,
        extra_predictors_hists,
    )

    chart = (
        (
            (
                alt.vconcat(
                    alt.vconcat(top_chart, time_line, spacing=130),
                    (
                        index_variables_hists
                        | c_0_hist
                        | error_hist & extra_predictors_hists
                    ),
                    spacing=10,
                )
            )
            .resolve_scale(color="independent")
            .add_selection(error_max_selection)
            .add_selection(c_0_min_selection)
            .add_selection(t_0_selection)
        )
        .configure_axis(labelFontSize=labelFontSize, titleFontSize=titleFontSize)
        .configure_legend(labelFontSize=labelFontSize)
    )
    if fitted_model.name == "RectangleFunction":
        chart = chart.add_selection(t_1_selection)

    vis_threshold = "id" + uuid.uuid4().hex
    new_sliders_threshold_id = "id" + uuid.uuid4().hex
    old_sliders_threshold_id = "id" + uuid.uuid4().hex
    new_index_threshold_id = "id" + uuid.uuid4().hex
    old_index_threshold_id = "id" + uuid.uuid4().hex
    html_chart = f"""
        <!DOCTYPE html>
        <html>
        <head>
          <script src="https://cdn.jsdelivr.net/npm/vega@{alt.VEGA_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-lite@{alt.VEGALITE_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-embed@{alt.VEGAEMBED_VERSION}"></script>
        </head>
        <body>

        <div class="container">
          <div class="row">
            <div>
            <div id={vis_threshold}></div>
            </div>
            <div style="position:absolute;left:920px;top:520px;width: -webkit-fill-available;">
            <div id={new_sliders_threshold_id}>
              <div id={old_sliders_threshold_id}></div>
            </div>
            </div>
            <div style="position:absolute;left:45px;top:390px;width: -webkit-fill-available;">
            <div id={new_index_threshold_id}>
              <div id={old_index_threshold_id}></div>
            </div>
            <hr/>
            <h1 style="text-align:center"> Interactive filters </h1>
            </div>
          </div>
        </div>

        <script type="text/javascript">
        vegaEmbed('#{vis_threshold}', {chart.to_json(indent=None)}).then(function(result) {{
            const sliders = document.getElementsByClassName('vega-bindings');
            const newestimate = document.getElementById('{new_sliders_threshold_id}');
            const oldestimate = document.getElementById('{old_sliders_threshold_id}');
            const newparent = document.getElementById('{new_index_threshold_id}');
            const oldchild = document.getElementById('{old_index_threshold_id}');
            for (var i = 0; i < sliders.length; i++) {{
                if (sliders[i].parentElement.parentElement.id == '{vis_threshold}') {{
                    var estimate_slider = sliders[i]
                    var index_slider = estimate_slider.querySelectorAll(".vega-bind")
                    }}
                }}
            newestimate.replaceChild(estimate_slider, oldestimate);
            for (var i = 0; i < index_slider.length; i++) {{
                if (index_slider[i].firstChild.innerHTML == "Plot average completeness per: ") {{
                    var index_color = index_slider[i]}}
                }}
            newparent.replaceChild(index_color, oldchild);
            }}).catch(console.error);
        </script>
        </body>
        </html>
        """
    display(HTML(html_chart))
    if save_path:
        save_html(
            obj=html_chart,
            filename=save_path,
        )



from datetime import datetime
from typing import List

import altair as alt

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.viz.plots.plot_probe.fitted_probe import fitted_probe_line
from edsteva.viz.plots.plot_probe.probe import probe_line
from edsteva.viz.utils import filter_predictor, save_html


def plot_probe(
    probe: BaseProbe,
    fitted_model: BaseModel = None,
    care_site_level: str = None,
    stay_type: List[str] = None,
    care_site_id: List[int] = None,
    start_date: datetime = None,
    end_date: datetime = None,
    care_site_short_name: List[int] = None,
    save_path: str = None,
    x_axis_title: str = "Time (Month Year)",
    x_grid: bool = True,
    y_axis_title: str = "Completeness predictor c(t)",
    y_grid: bool = True,
    show_n_visit: bool = False,
    show_per_care_site: bool = True,
    labelAngle: float = -90,
    labelFontSize: float = 12,
    titleFontSize: float = 13,
    **kwargs,
):
    r"""
    Displays a chart with the average completeness predictor $c(t)$ over time $t$ with the fitted model $\hat{c}(t)$ if specified.
    The chart is exportable in png or svg format and easy to integrate into a report. Is also possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$.
    fitted_model : BaseModel, optional
        Model fitted to the probe
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    stay_type : List[str], optional
        **EXAMPLE**: `"All"` or `["All", "Urg"]`
    care_site_id : List[int], optional
        **EXAMPLE**: `[8312056386, 8312027648]`
    start_date : datetime, optional
        **EXAMPLE**: `"2019-05-01"`
    end_date : datetime, optional
        **EXAMPLE**: `"2021-07-01"`
    care_site_short_name : List[int], optional
        **EXAMPLE**: `"HOSPITAL XXXX"`
    save_path : str, optional
        Folder path where to save the chart in HTML format.

        **EXAMPLE**: `"my_folder/my_file.html"`
    x_axis_title: str, optional,
        Label name for the x axis.
    x_grid: bool, optional,
        If False, remove the grid for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    y_grid: bool, optional,
        If False, remove the grid for the y axis.
    show_n_visit: bool, optional
        If True, compute the sum of the number of visit instead of the mean of the completeness predictor $c(t)$.
    show_per_care_site: bool, optional
        If True, the average completeness predictor $c(t)$ is computed for each care site independently. If False, it is computed over all care sites.
    labelAngle: float, optional
        The rotation angle of the label on the x_axis.
    labelFontSize: float, optional
        The font size of the labels (axis and legend).
    titleFontSize: float, optional
        The font size of the titles.
    """

    index = list(set(probe._index).difference(["care_site_level", "care_site_id"]))
    if show_per_care_site:
        index = index + ["care_site_short_name"]

    if fitted_model:
        predictor = fitted_model.predict(probe)
    else:
        predictor = probe.predictor

    predictor = filter_predictor(
        predictor=predictor,
        care_site_level=care_site_level,
        stay_type=stay_type,
        care_site_id=care_site_id,
        care_site_short_name=care_site_short_name,
        start_date=start_date,
        end_date=end_date,
        **kwargs,
    )

    index = [variable for variable in index if len(predictor[variable].unique()) >= 2]

    alt.data_transformers.disable_max_rows()

    if fitted_model:
        chart = fitted_probe_line(
            predictor=predictor,
            index=index,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            x_grid=x_grid,
            y_grid=y_grid,
            labelAngle=labelAngle,
            labelFontSize=labelFontSize,
        )
    else:
        chart = probe_line(
            predictor=predictor,
            index=index,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            x_grid=x_grid,
            y_grid=y_grid,
            show_n_visit=show_n_visit,
            labelAngle=labelAngle,
        )

    if save_path:
        save_html(
            obj=chart.configure_axis(
                labelFontSize=labelFontSize, titleFontSize=titleFontSize
            ).configure_legend(labelFontSize=labelFontSize),
            filename=save_path,
        )

    return chart



from typing import Callable, List

import pandas as pd

from edsteva.metrics import error_after_t0
from edsteva.models import BaseModel
from edsteva.models.step_function import algos


class StepFunction(BaseModel):
    r"""It models the completeness predictor $c(t)$ as a step function $f_{t_0, c_0}(t)$ as follow:

    $$
    f_{t_0, c_0}(t) = c_0 \ \mathbb{1}_{t \geq t_0}(t)
    $$

    It computes the following estimates $(t_0, c_0)$:

    - the characteristic time $t_0$ estimates the time after which the data is available
    - the characteristic value $c_0$ estimates the stabilized routine completeness

    Attributes
    ----------
    _coefs: List[str]
        Model coefficients

        **VALUE**: ``["t_0", "c_0"]``

    Example
    ----------

    ```python
    from edsteva.models.step_function import StepFunction

    step_function_model = StepFunction()
    step_function_model.fit(probe)
    step_function_model.estimates.head()
    ```

    | care_site_level          | care_site_id | stay_type    | t_0        | c_0   |
    | :----------------------- | :----------- | :----------- | :--------- | :---- |
    | Unité Fonctionnelle (UF) | 8312056386   | 'Urg_Hospit' | 2019-05-01 | 0.397 |
    | Unité Fonctionnelle (UF) | 8312056386   | 'All'        | 2011-04-01 | 0.583 |
    | Pôle/DMU                 | 8312027648   | 'Urg_Hospit' | 2021-03-01 | 0.677 |
    | Pôle/DMU                 | 8312027648   | 'All'        | 2018-08-01 | 0.764 |
    | Hôpital                  | 8312022130   | 'Urg_Hospit' | 2022-02-01 | 0.652 |
    """

    _coefs = ["t_0", "c_0"]

    def fit_process(
        self,
        predictor: pd.DataFrame,
        index: List[str] = None,
        algo: Callable = algos.loss_minimization,
        **kwargs
    ) -> None:
        """Script to be used by [``fit()``][edsteva.models.base.BaseModel.fit]

        Parameters
        ----------
        predictor : pd.DataFrame
            Target variable to be fitted
        index : List[str], optional
            Variable from which data is grouped

            **EXAMPLE**: `["care_site_level", "stay_type", "note_type", "care_site_id"]`
        algo : Callable, optional
            Algorithm used for the coefficients estimation ($t_0$ and $c_0$)
        """

        estimates = algo(predictor=predictor, index=index, **kwargs)

        return estimates[index + self._coefs]

    def predict_process(
        self,
        predictor: pd.DataFrame,
        index: List[str],
    ) -> pd.DataFrame:
        """Script to be used by [``predict()``][edsteva.models.base.BaseModel.predict]

        Parameters
        ----------
        predictor : pd.DataFrame
            Target DataFrame to be predicted
        index : List[str]
            List of the columns given by Probe._index

        Returns
        -------
        pd.DataFrame
            Prediction

        Raises
        ------
        Exception
            Some indexes have no associated estimates, the model must be fitted on an adequate probe

        Examples
        --------
        | care_site_level          | care_site_id | care_site_short_name | stay_type    | date       | n_visit | c     | c_fit |
        | :----------------------- | :----------- | :------------------- | :----------- | :--------- | :------ | :---- | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'Urg_Hospit' | 2019-05-01 | 233.0   | 0.841 | 0.758 |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'All'        | 2021-04-01 | 393.0   | 0.640 | 0.758 |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'Urg_Hospit' | 2011-03-01 | 204.0   | 0.497 | 0     |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'All'        | 2018-08-01 | 22.0    | 0.784 | 0.874 |
        | Hôpital                  | 8312022130   | Care site 3          | 'Urg_Hospit' | 2022-02-01 | 9746.0  | 0.974 | 0.912 |
        """

        prediction = self.is_predictable_probe(predictor=predictor, index=index)

        prediction["c_hat"] = prediction["c_0"].where(
            prediction["date"] >= prediction["t_0"], 0
        )
        return prediction.drop(columns=self._coefs + self._metrics)

    def default_metrics(
        self,
        predictor: pd.DataFrame,
        estimates: pd.DataFrame,
        index: List[str],
    ):
        r"""Default metrics used if metric_functions is set to None. Here the default metric is the mean squared error computed after $t_0$

        Parameters
        ----------
        predictor : pd.DataFrame
            Target DataFrame describing the completeness predictor $c(t)$
        estimates : pd.DataFrame
            Target DataFrame describing the estimates $(\hat{t_0}, \hat{c_0})$
        index : List[str]
            Variable from which data is grouped

            **EXAMPLE**: `["care_site_level", "stay_type", "note_type", "care_site_id"]`
        """
        return error_after_t0(predictor=predictor, estimates=estimates, index=index)



import os
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union

import _pickle as pickle
import pandas as pd
from IPython.display import display
from loguru import logger

from edsteva.utils.checks import check_columns
from edsteva.utils.framework import get_framework, to
from edsteva.utils.typing import Data, DataFrame

CARE_SITE_LEVEL_NAMES = {
    "Hospital": "Hôpital",
    "Pole": "Pôle/DMU",
    "UF": "Unité Fonctionnelle (UF)",
}

UNSUPPORTED_CARE_SITE_LEVEL_NAMES = {
    "UC": "Unité de consultation (UC)",
}


def prepare_visit_occurrence(data, start_date, end_date, stay_types):
    check_columns(
        data.visit_occurrence,
        required_columns=[
            "visit_occurrence_id",
            "visit_source_value",
            "visit_start_datetime",
            "care_site_id",
            "row_status_source_value",
            "visit_occurrence_source_value",
        ],
        df_name="visit_occurrence",
    )
    visit_occurrence = data.visit_occurrence[
        [
            "visit_occurrence_id",
            "visit_source_value",
            "visit_start_datetime",
            "care_site_id",
            "row_status_source_value",
            "visit_occurrence_source_value",
        ]
    ]
    visit_occurrence = visit_occurrence.rename(
        columns={"visit_source_value": "stay_type", "visit_start_datetime": "date"}
    )
    visit_occurrence = get_valid_observations(
        table=visit_occurrence,
        table_name="visit_occurrence",
        invalid_naming="supprimé",
    )
    visit_occurrence = filter_table_by_date(
        table=visit_occurrence,
        table_name="visit_occurrence",
        start_date=start_date,
        end_date=end_date,
    )

    if stay_types:
        visit_occurrence = filter_table_by_type(
            table=visit_occurrence,
            table_name="visit_occurrence",
            type_groups=stay_types,
            name="stay_type",
        )

    return visit_occurrence


def prepare_condition_occurrence(
    data: Data,
    extra_data: Data,
    source: str,
    diag_types: List[str],
    condition_types: Dict,
):
    if source == "AREM":
        # Fetch conditions from Data lake
        I2B2_visit = extra_data.visit_occurrence[
            ["visit_occurrence_id", "visit_occurrence_source_value"]
        ]
        I2B2_condition_occurrence = extra_data.condition_occurrence[
            [
                "visit_occurrence_id",
                "condition_status_source_value",
                "condition_source_value",
                "cdm_source",
            ]
        ]
        # Add visit_occurrence_source_value
        condition_occurrence = I2B2_visit.merge(
            I2B2_condition_occurrence,
            on="visit_occurrence_id",
            how="inner",
        ).drop(columns="visit_occurrence_id")

    else:
        condition_occurrence = data.condition_occurrence[
            [
                "visit_occurrence_id",
                "condition_source_value",
                "condition_status_source_value",
                "row_status_source_value",
                "cdm_source",
            ]
        ]
        condition_occurrence = get_valid_observations(
            table=condition_occurrence,
            table_name="condition_occurrence",
            valid_naming="Actif",
        )

    # Keep source observations
    condition_occurrence = condition_occurrence[
        condition_occurrence.cdm_source == source
    ].drop(columns="cdm_source")

    # Filter diagnostics
    condition_occurrence = condition_occurrence.rename(
        columns={"condition_status_source_value": "diag_type"}
    )
    if diag_types:
        condition_occurrence = filter_table_by_type(
            table=condition_occurrence,
            table_name="condition_occurrence",
            type_groups=diag_types,
            name="diag_type",
        )

    # Filter conditions
    condition_occurrence = condition_occurrence.rename(
        columns={"condition_source_value": "condition_type"}
    )
    if condition_types:
        condition_occurrence = filter_table_by_type(
            table=condition_occurrence,
            table_name="condition_occurrence",
            type_groups=condition_types,
            name="condition_type",
        )

    return condition_occurrence


def prepare_care_site(
    data,
    care_site_ids,
    care_site_short_names,
    care_site_relationship,
):
    care_site = data.care_site[
        [
            "care_site_id",
            "care_site_type_source_value",
            "care_site_short_name",
        ]
    ]
    care_site = care_site.rename(
        columns={"care_site_type_source_value": "care_site_level"}
    )
    if care_site_ids or care_site_short_names:
        care_site = filter_table_by_care_site(
            table_to_filter=care_site,
            table_name="care_site",
            care_site_relationship=care_site_relationship,
            care_site_ids=care_site_ids,
            care_site_short_names=care_site_short_names,
        )

    return care_site


def prepare_note(data, note_types):
    note = data.note[
        [
            "note_id",
            "visit_occurrence_id",
            "note_class_source_value",
            "row_status_source_value",
            "note_text",
        ]
    ]
    note = note.rename(columns={"note_class_source_value": "note_type"})
    note = note[~(note["note_text"].isna())]
    note = note.drop(columns=["note_text"])
    note = get_valid_observations(table=note, table_name="note", valid_naming="Actif")
    if note_types:
        note = filter_table_by_type(
            table=note,
            table_name="note",
            type_groups=note_types,
            name="note_type",
        )

    return note


def prepare_visit_detail(data, start_date, end_date):
    visit_detail = data.visit_detail[
        [
            "visit_detail_id",
            "visit_occurrence_id",
            "visit_detail_start_datetime",
            "visit_detail_type_source_value",
            "care_site_id",
            "row_status_source_value",
        ]
    ]
    visit_detail = visit_detail.rename(
        columns={
            "visit_detail_id": "visit_id",
            "visit_detail_start_datetime": "date",
        }
    )
    visit_detail = get_valid_observations(
        table=visit_detail, table_name="visit_detail", valid_naming="Actif"
    )
    visit_detail = visit_detail[
        visit_detail["visit_detail_type_source_value"] == "PASS"
    ]  # Important to filter only "PASS" to remove duplicate visits
    visit_detail = visit_detail.drop(columns=["visit_detail_type_source_value"])
    visit_detail = filter_table_by_date(
        table=visit_detail,
        table_name="visit_detail",
        start_date=start_date,
        end_date=end_date,
    )

    return visit_detail


def hospital_only(care_site_levels: List[str]):
    if not isinstance(care_site_levels, list):
        care_site_levels = [care_site_levels]
    return len(care_site_levels) == 1 and (
        care_site_levels[0] == "Hospital"
        or care_site_levels[0] == CARE_SITE_LEVEL_NAMES["Hospital"]
    )


def filter_table_by_care_site(
    table_to_filter: DataFrame,
    table_name: str,
    care_site_relationship: pd.DataFrame,
    care_site_short_names: Union[str, List[str]] = None,
    care_site_ids: Union[int, List[int]] = None,
):
    care_site = care_site_relationship[
        ["care_site_id", "care_site_short_name", "care_site_level"]
    ]
    care_site_filter = []
    if care_site_ids:
        if not isinstance(care_site_ids, list):
            care_site_ids = [care_site_ids]
        care_site_filter.append(
            care_site[care_site["care_site_id"].isin(care_site_ids)].copy()
        )
    if care_site_short_names:
        if not isinstance(care_site_short_names, list):
            care_site_short_names = [care_site_short_names]
        care_site_filter.append(
            care_site[
                care_site["care_site_short_name"].isin(care_site_short_names)
            ].copy()
        )
    if care_site_filter:
        care_site_filter = pd.concat(
            care_site_filter, ignore_index=True
        ).drop_duplicates("care_site_id")
    else:
        raise Exception("care_site_ids or care_site_short_names must be provided")

    # Get all UF from UC
    uc_care_site = care_site_filter[
        care_site_filter.care_site_level == UNSUPPORTED_CARE_SITE_LEVEL_NAMES["UC"]
    ]
    uc_care_site = uc_care_site[["care_site_id"]].drop_duplicates()
    care_site_rel_uc_to_uf = _get_relationship_table_uc_to_uf(
        care_site_relationship=care_site_relationship
    )
    care_site_rel_uc_to_uf = to("pandas", care_site_rel_uc_to_uf)
    related_uf_care_site = uc_care_site.merge(
        care_site_rel_uc_to_uf,
        on="care_site_id",
    )
    uf_from_uc = related_uf_care_site[
        ["care_site_id_uf", "care_site_short_name_uf"]
    ].drop_duplicates("care_site_id_uf")

    # Get all UF from Hospital
    care_site_rel_uf_to_hospital = _get_relationship_table_uf_to_hospital(
        care_site_relationship=care_site_relationship
    )
    care_site_rel_uf_to_hospital = to("pandas", care_site_rel_uf_to_hospital)

    hospital_care_site = care_site_filter[
        care_site_filter.care_site_level == CARE_SITE_LEVEL_NAMES["Hospital"]
    ].rename(
        columns={
            "care_site_id": "care_site_id_hospital",
            "care_site_short_name": "care_site_short_name_hospital",
        }
    )
    hospital_care_site = hospital_care_site[
        ["care_site_id_hospital", "care_site_short_name_hospital"]
    ].drop_duplicates("care_site_id_hospital")
    related_hospital_care_site = hospital_care_site.merge(
        care_site_rel_uf_to_hospital,
        on=[
            "care_site_id_hospital",
            "care_site_short_name_hospital",
        ],
        how="left",
    )
    uf_from_hospital = related_hospital_care_site[
        [
            "care_site_id_uf",
            "care_site_short_name_uf",
        ]
    ].drop_duplicates("care_site_id_uf")
    pole_from_hospital = related_hospital_care_site[
        [
            "care_site_id_pole",
            "care_site_short_name_pole",
        ]
    ].drop_duplicates("care_site_id_pole")

    # Get all UF from pole
    pole_care_site = care_site_filter[
        care_site_filter.care_site_level == CARE_SITE_LEVEL_NAMES["Pole"]
    ].rename(
        columns={
            "care_site_id": "care_site_id_pole",
            "care_site_short_name": "care_site_short_name_pole",
        }
    )
    pole_care_site = pole_care_site[
        ["care_site_id_pole", "care_site_short_name_pole"]
    ].drop_duplicates("care_site_id_pole")
    related_pole_care_site = pole_care_site.merge(
        care_site_rel_uf_to_hospital,
        on=[
            "care_site_id_pole",
            "care_site_short_name_pole",
        ],
        how="left",
    )
    uf_from_pole = related_pole_care_site[
        [
            "care_site_id_uf",
            "care_site_short_name_uf",
        ]
    ].drop_duplicates("care_site_id_uf")

    # Get all Hospital from Pole
    hospital_from_pole = related_pole_care_site[
        [
            "care_site_id_hospital",
            "care_site_short_name_hospital",
        ]
    ].drop_duplicates("care_site_id_hospital")

    # Get all Pole from UF
    uf_care_site = care_site_filter[
        care_site_filter.care_site_level == CARE_SITE_LEVEL_NAMES["UF"]
    ].rename(
        columns={
            "care_site_id": "care_site_id_uf",
            "care_site_short_name": "care_site_short_name_uf",
        }
    )
    uf_care_site = uf_care_site[
        ["care_site_id_uf", "care_site_short_name_uf"]
    ].drop_duplicates("care_site_id_uf")
    uf_care_site = pd.concat([uf_care_site, uf_from_uc])

    related_uf_care_site = uf_care_site.merge(
        care_site_rel_uf_to_hospital,
        on=[
            "care_site_id_uf",
            "care_site_short_name_uf",
        ],
        how="left",
    )
    pole_from_uf = related_uf_care_site[
        [
            "care_site_id_pole",
            "care_site_short_name_pole",
        ]
    ].drop_duplicates("care_site_id_pole")

    # Get all Hospital from UF
    hospital_from_uf = related_uf_care_site[
        [
            "care_site_id_hospital",
            "care_site_short_name_hospital",
        ]
    ].drop_duplicates("care_site_id_hospital")

    extended_uf_care_site = pd.concat(
        [uf_care_site, uf_from_pole, uf_from_hospital, uf_from_uc], ignore_index=True
    ).drop_duplicates("care_site_id_uf")
    extended_uf_care_site = extended_uf_care_site.rename(
        columns={
            "care_site_id_uf": "care_site_id",
            "care_site_short_name_uf": "care_site_short_name",
        }
    )

    extended_pole_care_site = pd.concat(
        [pole_care_site, pole_from_uf, pole_from_hospital], ignore_index=True
    ).drop_duplicates("care_site_id_pole")
    extended_pole_care_site = extended_pole_care_site.rename(
        columns={
            "care_site_id_pole": "care_site_id",
            "care_site_short_name_pole": "care_site_short_name",
        }
    )
    extended_hospital_care_site = pd.concat(
        [hospital_care_site, hospital_from_pole, hospital_from_uf], ignore_index=True
    ).drop_duplicates("care_site_id_hospital")
    extended_hospital_care_site = extended_hospital_care_site.rename(
        columns={
            "care_site_id_hospital": "care_site_id",
            "care_site_short_name_hospital": "care_site_short_name",
        }
    )
    unsupported_care_site = care_site_filter[
        ~(care_site_filter.care_site_level.isin(CARE_SITE_LEVEL_NAMES.values()))
    ]
    if not unsupported_care_site.empty:
        logger.warning(
            "The following care site ids are not supported because the associated care site levels are not in {}.",
            CARE_SITE_LEVEL_NAMES.values(),
        )
        display(unsupported_care_site)

    if not extended_hospital_care_site.empty:
        logger.debug(
            "The following hospitals {} have been selected from {}.",
            extended_hospital_care_site.care_site_short_name.to_list(),
            table_name,
        )
    if not extended_pole_care_site.empty:
        logger.debug(
            "The following poles {} have been selected from {}.",
            extended_pole_care_site.care_site_short_name.to_list(),
            table_name,
        )
    if not extended_uf_care_site.empty:
        logger.debug(
            "The following UF {} have been selected from {}.",
            extended_uf_care_site.care_site_short_name.to_list(),
            table_name,
        )

    extended_care_site_id_to_filter = pd.concat(
        [extended_hospital_care_site, extended_pole_care_site, extended_uf_care_site],
        ignore_index=True,
    ).care_site_id.to_list()
    return table_to_filter[
        table_to_filter["care_site_id"].isin(extended_care_site_id_to_filter)
    ]


def filter_table_by_type(
    table: DataFrame,
    table_name: str,
    type_groups: Union[str, Dict],
    name: str,
):
    if isinstance(type_groups, str):
        type_groups = {type_groups: type_groups}
    if isinstance(type_groups, dict):
        table_per_types = []
        for type_name, type_value in type_groups.items():
            table_per_type_element = table[
                table[name].str.contains(
                    type_value,
                    case=False,
                    regex=True,
                    na=False,
                )
            ].copy()
            table_per_type_element[name] = type_name
            table_per_types.append(table_per_type_element)
    else:
        raise TypeError("{} must be str or dict not {}".format(name, type(type_groups)))

    logger.debug(
        "The following {} : {} have been selected on table {}",
        name,
        type_groups,
        table_name,
    )
    return get_framework(table).concat(table_per_types, ignore_index=True)


def convert_table_to_uf(
    table: DataFrame,
    table_name: str,
    care_site_relationship: DataFrame,
):
    care_site_rel_uc_to_uf = _get_relationship_table_uc_to_uf(
        care_site_relationship=care_site_relationship
    )
    care_site_rel_uc_to_uf = to(get_framework(table), care_site_rel_uc_to_uf)
    table = table.merge(care_site_rel_uc_to_uf, on="care_site_id", how="left")
    table["care_site_id"] = table["care_site_id_uf"].mask(
        table["care_site_id_uf"].isna(), table["care_site_id"]
    )
    table = table.drop(columns=["care_site_id_uf"])
    logger.debug(
        "For level {}, stays of the table {} located in {} have been linked to their corresponding {}",
        CARE_SITE_LEVEL_NAMES["UF"],
        table_name,
        UNSUPPORTED_CARE_SITE_LEVEL_NAMES["UC"],
        CARE_SITE_LEVEL_NAMES["UF"],
    )
    return table


def convert_table_to_pole(
    table: DataFrame,
    table_name: str,
    care_site_relationship: DataFrame,
):
    care_site_rel_uf_to_pole = _get_relationship_table_uf_to_hospital(
        care_site_relationship=care_site_relationship
    )[["care_site_id_uf", "care_site_id_pole"]].rename(
        columns={"care_site_id_uf": "care_site_id"}
    )
    care_site_rel_uf_to_pole = to(get_framework(table), care_site_rel_uf_to_pole)
    table = table.merge(care_site_rel_uf_to_pole, on="care_site_id", how="left")
    table["care_site_id"] = table["care_site_id_pole"].mask(
        table["care_site_id_pole"].isna(), table["care_site_id"]
    )
    table = table.drop(columns="care_site_id_pole")
    logger.debug(
        "For level {}, stays of the table {} located in {} have been linked to their corresponding {}",
        CARE_SITE_LEVEL_NAMES["Pole"],
        table_name,
        CARE_SITE_LEVEL_NAMES["UF"],
        CARE_SITE_LEVEL_NAMES["Pole"],
    )
    return table


def get_valid_observations(
    table: DataFrame,
    table_name: str,
    invalid_naming: str = None,
    valid_naming: str = None,
):
    check_columns(
        df=table,
        required_columns=["row_status_source_value"],
    )

    if valid_naming:
        table_valid = table[table["row_status_source_value"] == valid_naming]
    elif invalid_naming:
        table_valid = table[~(table["row_status_source_value"] == invalid_naming)]
    else:
        raise Exception("valid_naming or invalid_naming must be provided.")
    table_valid = table_valid.drop(columns=["row_status_source_value"])
    logger.debug("Valid observations have been selected for table {}.", table_name)
    return table_valid


def filter_table_by_date(
    table: DataFrame,
    table_name: str,
    start_date: Union[datetime, str] = None,
    end_date: Union[datetime, str] = None,
):

    check_columns(df=table, required_columns=["date"])

    table.dropna(subset=["date"], inplace=True)
    logger.debug("Droping observations with missing date in table {}.", table_name)
    table["date"] = table["date"].astype("datetime64")
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)

    if end_date and start_date:
        table = table[(table["date"] >= start_date) & (table["date"] < end_date)]
        logger.debug(
            "Observations between {} and {} have been selected for table {}.",
            start_date,
            end_date,
            table_name,
        )
    elif start_date:
        table = table[table["date"] >= start_date]
        logger.debug(
            "Observations after {} have been selected for table {}.",
            start_date,
            table_name,
        )
    elif end_date:
        table = table[table["date"] < end_date]
        logger.debug(
            "Observations before {} have been selected for table {}.",
            end_date,
            table_name,
        )
    # Truncate
    table["date"] = table["date"].dt.strftime("%Y-%m").astype("datetime64")
    return table


def save_object(obj, filename):
    if not isinstance(filename, Path):
        filename = Path(filename)
    os.makedirs(filename.parent, exist_ok=True)
    with open(filename, "wb") as outp:  # Overwrites any existing file.
        pickle.dump(obj, outp, -1)
        logger.info("Saved to {}", filename)


def load_object(filename):
    if os.path.isfile(filename):
        with open(filename, "rb") as obj:
            logger.info("Successfully loaded from {}", filename)
            return pickle.load(obj)
    else:
        raise FileNotFoundError(
            "There is no file found in {}".format(
                filename,
            )
        )


def delete_object(obj, filename: str):
    if os.path.isfile(filename):
        os.remove(filename)
        logger.info(
            "Removed from {}",
            filename,
        )
        if hasattr(obj, "path"):
            del obj.path
    else:
        logger.warning(
            "There is no file found in {}",
            filename,
        )


def _get_relationship_table_uc_to_uf(
    care_site_relationship: DataFrame,
):
    check_columns(
        df=care_site_relationship,
        required_columns=[
            "care_site_id",
            "care_site_level",
            "care_site_short_name",
            "parent_care_site_id",
            "parent_care_site_level",
            "parent_care_site_short_name",
        ],
    )

    care_site_relationship = care_site_relationship[
        [
            "care_site_id",
            "care_site_level",
            "care_site_short_name",
            "parent_care_site_id",
            "parent_care_site_level",
            "parent_care_site_short_name",
        ]
    ]

    uc_care_site = care_site_relationship[
        (
            care_site_relationship["care_site_level"]
            == UNSUPPORTED_CARE_SITE_LEVEL_NAMES["UC"]
        )
    ]

    care_site_rel_grandparent = care_site_relationship[
        [
            "care_site_id",
            "parent_care_site_id",
            "parent_care_site_short_name",
            "parent_care_site_level",
        ]
    ].rename(
        columns={
            "care_site_id": "parent_care_site_id",
            "parent_care_site_id": "grandparent_care_site_id",
            "parent_care_site_short_name": "grandparent_care_site_short_name",
            "parent_care_site_level": "grandparent_care_site_level",
        }
    )
    uc_care_site = uc_care_site.merge(
        care_site_rel_grandparent, on="parent_care_site_id"
    )
    uc_care_site["care_site_id_uf"] = None
    uc_care_site["care_site_id_uf"] = uc_care_site["care_site_id_uf"].mask(
        uc_care_site["grandparent_care_site_level"] == CARE_SITE_LEVEL_NAMES["UF"],
        uc_care_site["grandparent_care_site_id"],
    )
    uc_care_site["care_site_id_uf"] = uc_care_site["care_site_id_uf"].mask(
        uc_care_site["parent_care_site_level"] == CARE_SITE_LEVEL_NAMES["UF"],
        uc_care_site["parent_care_site_id"],
    )
    uc_care_site["care_site_short_name_uf"] = None
    uc_care_site["care_site_short_name_uf"] = uc_care_site[
        "care_site_short_name_uf"
    ].mask(
        uc_care_site["grandparent_care_site_level"] == CARE_SITE_LEVEL_NAMES["UF"],
        uc_care_site["care_site_short_name"],
    )
    uc_care_site["care_site_short_name_uf"] = uc_care_site[
        "care_site_short_name_uf"
    ].mask(
        uc_care_site["parent_care_site_level"] == CARE_SITE_LEVEL_NAMES["UF"],
        uc_care_site["care_site_short_name"],
    )
    return uc_care_site[
        [
            "care_site_id",
            "care_site_id_uf",
            "care_site_short_name_uf",
        ]
    ].drop_duplicates(["care_site_id", "care_site_id_uf"])


def _get_relationship_table_uf_to_hospital(
    care_site_relationship: DataFrame,
):
    check_columns(
        df=care_site_relationship,
        required_columns=[
            "care_site_id",
            "care_site_level",
            "care_site_short_name",
            "parent_care_site_id",
            "parent_care_site_level",
            "parent_care_site_short_name",
        ],
    )

    care_site_rel_uf_pole = care_site_relationship.rename(
        columns={
            "care_site_level": "care_site_level_uf",
            "care_site_id": "care_site_id_uf",
            "care_site_short_name": "care_site_short_name_uf",
            "parent_care_site_level": "care_site_level_pole",
            "parent_care_site_id": "care_site_id_pole",
            "parent_care_site_short_name": "care_site_short_name_pole",
        }
    )

    uf_to_pole_care_site = care_site_rel_uf_pole[
        (care_site_rel_uf_pole["care_site_level_uf"] == CARE_SITE_LEVEL_NAMES["UF"])
    ]

    care_site_rel_pole_hospit = care_site_relationship.rename(
        columns={
            "care_site_id": "care_site_id_pole",
            "parent_care_site_level": "care_site_level_hospital",
            "parent_care_site_id": "care_site_id_hospital",
            "parent_care_site_short_name": "care_site_short_name_hospital",
        }
    ).drop(columns=["care_site_level", "care_site_short_name"])

    uf_to_hospital_care_site = uf_to_pole_care_site.merge(
        care_site_rel_pole_hospit, on="care_site_id_pole"
    )

    uf_to_hospital_care_site = uf_to_hospital_care_site[
        uf_to_hospital_care_site.care_site_level_pole == CARE_SITE_LEVEL_NAMES["Pole"]
    ]
    uf_to_hospital_care_site = uf_to_hospital_care_site[
        uf_to_hospital_care_site.care_site_level_hospital
        == CARE_SITE_LEVEL_NAMES["Hospital"]
    ]
    uf_to_hospital_care_site = uf_to_hospital_care_site.drop_duplicates(
        [
            "care_site_id_uf",
            "care_site_id_pole",
            "care_site_id_hospital",
        ]
    )
    return uf_to_hospital_care_site[
        [
            "care_site_id_uf",
            "care_site_short_name_uf",
            "care_site_id_pole",
            "care_site_short_name_pole",
            "care_site_id_hospital",
            "care_site_short_name_hospital",
        ]
    ]


def get_care_site_relationship(data: Data) -> pd.DataFrame:
    """Computes hierarchical care site structure

    Parameters
    ----------
    data : Data
        Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]

    Example
    -------

    | care_site_id | care_site_level            | care_site_short_name | parent_care_site_id | parent_care_site_level     | parent_care_site_short_name |
    | :----------- | :------------------------- | :------------------- | :------------------ | :------------------------- | :-------------------------- |
    | 8312056386   | Unité Fonctionnelle (UF)   | UF A                 | 8312027648          | Pôle/DMU                   | Pole A                      |
    | 8312022130   | Pôle/DMU                   | Pole B               | 8312033550          | Hôpital                    | Hospital A                  |
    | 8312016782   | Service/Département        | Service A            | 8312033550          | Hôpital                    | Hospital A                  |
    | 8312010155   | Unité Fonctionnelle (UF)   | UF B                 | 8312022130          | Pôle/DMU                   | Pole B                      |
    | 8312067829   | Unité de consultation (UC) | UC A                 | 8312051097          | Unité de consultation (UC) | UC B                        |

    """
    fact_relationship = data.fact_relationship[
        [
            "fact_id_1",
            "fact_id_2",
            "domain_concept_id_1",
            "relationship_concept_id",
        ]
    ]
    fact_relationship = to("pandas", fact_relationship)

    care_site_relationship = fact_relationship[
        (fact_relationship["domain_concept_id_1"] == 57)  # Care_site domain
        & (fact_relationship["relationship_concept_id"] == 46233688)  # Included in
    ]
    care_site_relationship = care_site_relationship.drop(
        columns=["domain_concept_id_1", "relationship_concept_id"]
    )
    care_site_relationship = care_site_relationship.rename(
        columns={"fact_id_1": "care_site_id", "fact_id_2": "parent_care_site_id"}
    )

    care_site = data.care_site[
        [
            "care_site_id",
            "care_site_type_source_value",
            "care_site_short_name",
        ]
    ]
    care_site = to("pandas", care_site)
    care_site = care_site.rename(
        columns={
            "care_site_type_source_value": "care_site_level",
        }
    )
    care_site_relationship = care_site.merge(
        care_site_relationship, on="care_site_id", how="left"
    )

    parent_care_site = care_site.rename(
        columns={
            "care_site_level": "parent_care_site_level",
            "care_site_id": "parent_care_site_id",
            "care_site_short_name": "parent_care_site_short_name",
        }
    )
    logger.debug("Create care site relationship to link UC to UF and UF to Pole")

    return care_site_relationship.merge(
        parent_care_site, on="parent_care_site_id", how="left"
    )


def concatenate_predictor_by_level(
    predictor_by_level: Dict[str, DataFrame],
    care_site_levels: List[str] = None,
) -> DataFrame:
    predictors_to_concat = []
    if care_site_levels:
        if not isinstance(care_site_levels, list):
            care_site_levels = [care_site_levels]
        unknown_levels, selected_levels = [], []
        for level in care_site_levels:
            if level in predictor_by_level:
                predictors_to_concat.append(predictor_by_level[level])
                selected_levels.append(level)
            elif level in CARE_SITE_LEVEL_NAMES:
                predictors_to_concat.append(
                    predictor_by_level[CARE_SITE_LEVEL_NAMES[level]]
                )
                selected_levels.append(level)
            else:
                unknown_levels.append(level)

        logger.debug(
            "The following levels {} have been selected",
            selected_levels,
        )
        if unknown_levels:
            logger.warning(
                "Unrecognized levels {}.the only supported levels are: {}",
                unknown_levels,
                list(CARE_SITE_LEVEL_NAMES.values())
                + list(CARE_SITE_LEVEL_NAMES.keys()),
            )
    else:
        predictors_to_concat = list(predictor_by_level.values())
        logger.debug(
            "The following levels {} have been selected",
            list(predictor_by_level.keys()),
        )

    if not predictors_to_concat:
        raise AttributeError(
            "care site levels must include at least one of the following levels: {}".format(
                list(CARE_SITE_LEVEL_NAMES.values())
                + list(CARE_SITE_LEVEL_NAMES.keys())
            )
        )

    framework = get_framework(predictors_to_concat[0])
    return framework.concat(predictors_to_concat)



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd

from edsteva.probes.base import BaseProbe
from edsteva.probes.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    convert_table_to_pole,
    convert_table_to_uf,
    hospital_only,
    prepare_care_site,
    prepare_note,
    prepare_visit_detail,
    prepare_visit_occurrence,
)
from edsteva.utils.checks import check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data


def compute_completeness(note_predictor):
    partition_cols = [
        "care_site_level",
        "care_site_id",
        "care_site_short_name",
        "stay_type",
        "note_type",
        "date",
    ]
    n_visit_with_note = (
        note_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"has_note": "count"})
        .rename(columns={"has_note": "n_visit_with_note"})
    )

    partition_cols = list(set(partition_cols) - {"note_type"})
    n_visit = (
        note_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"visit_id": "nunique"})
        .rename(columns={"visit_id": "n_visit"})
    )

    note_predictor = n_visit_with_note.merge(
        n_visit,
        on=partition_cols,
    )

    note_predictor = to("pandas", note_predictor)

    note_predictor["c"] = note_predictor["n_visit"].where(
        note_predictor["n_visit"] == 0,
        note_predictor["n_visit_with_note"] / note_predictor["n_visit"],
    )
    note_predictor = note_predictor.drop(columns=["n_visit_with_note"])

    return note_predictor


def get_hospital_visit(note, visit_occurrence, care_site):
    note_hospital = note[["visit_occurrence_id", "note_type"]].drop_duplicates()
    note_hospital["has_note"] = True
    hospital_visit = visit_occurrence.merge(
        note_hospital, on="visit_occurrence_id", how="left"
    )
    hospital_visit = hospital_visit.rename(columns={"visit_occurrence_id": "visit_id"})
    hospital_visit = hospital_visit.merge(care_site, on="care_site_id")
    if is_koalas(hospital_visit):
        hospital_visit.spark.cache()

    return hospital_visit


def get_uf_visit(
    extra_data,
    note,
    visit_occurrence,
    visit_detail,
    care_site,
    care_site_relationship,
):  # pragma: no cover
    # Load Orbis note and Uf for Note
    note_orbis = extra_data.orbis_document[
        [
            "ids_eds",
            "id_dm_doc_ufr",
            "id_dm_doc_us",
        ]
    ]
    note_orbis = note_orbis.rename(columns={"ids_eds": "note_id"})

    orbis_care_site = extra_data.orbis_ref_struct_list[
        [
            "id_ref_stuct",
            "ids_eds",
        ]
    ]
    orbis_care_site = orbis_care_site.rename(
        columns={
            "id_ref_stuct": "id_orbis",
            "ids_eds": "care_site_id",
        }
    )

    note = note.merge(note_orbis, on="note_id")
    note = note.melt(
        id_vars=["note_id", "note_type", "visit_occurrence_id"],
        value_name="id_orbis",
    )
    note = note.merge(
        orbis_care_site,
        on="id_orbis",
    )

    note_uf = note[
        ["visit_occurrence_id", "note_type", "care_site_id"]
    ].drop_duplicates()
    note_uf["has_note"] = True

    visit_detail = visit_detail.merge(
        visit_occurrence[["visit_occurrence_id", "stay_type"]],
        on="visit_occurrence_id",
    )
    visit_detail = visit_detail.merge(
        note_uf,
        on=["visit_occurrence_id", "care_site_id"],
        how="left",
    )
    visit_detail = visit_detail.drop(columns="visit_occurrence_id")

    uf_visit = convert_table_to_uf(
        table=visit_detail,
        table_name="visit_detail",
        care_site_relationship=care_site_relationship,
    )
    uf_visit = uf_visit.merge(care_site, on="care_site_id")

    uf_name = CARE_SITE_LEVEL_NAMES["UF"]
    uf_visit = uf_visit[uf_visit["care_site_level"] == uf_name]

    if is_koalas(uf_visit):
        uf_visit.spark.cache()

    return uf_visit


def get_pole_visit(uf_visit, care_site, care_site_relationship):  # pragma: no cover

    pole_visit = convert_table_to_pole(
        table=uf_visit.drop(columns=["care_site_short_name", "care_site_level"]),
        table_name="uf_visit",
        care_site_relationship=care_site_relationship,
    )

    pole_visit = pole_visit.merge(care_site, on="care_site_id")

    pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
    pole_visit = pole_visit[pole_visit["care_site_level"] == pole_name]

    if is_koalas(pole_visit):
        pole_visit.spark.cache()

    return pole_visit


class NoteProbe(BaseProbe):
    r"""
    The ``NoteProbe`` computes $c(t)$ the availability of clinical documents linked to patients' administrative visit:

    $$
    c(t) = \frac{n_{with\,doc}(t)}{n_{visit}(t)}
    $$

    Where $n_{visit}(t)$ is the number of visits, $n_{with\,doc}$ the number of visits having at least one document and $t$ is the month.

    Attributes
    ----------
    _index: List[str]
        Variable from which data is grouped

        **VALUE**: ``["care_site_level", "stay_type", "note_type", "care_site_id"]``
    """

    _index = ["care_site_level", "stay_type", "note_type", "care_site_id"]

    def compute_process(
        self,
        data: Data,
        care_site_relationship: pd.DataFrame,
        extra_data: Data = None,
        start_date: datetime = None,
        end_date: datetime = None,
        care_site_levels: List[str] = None,
        care_site_short_names: List[str] = None,
        stay_types: Union[str, Dict[str, str]] = None,
        note_types: Union[str, Dict[str, str]] = {
            "All": ".*",
            "Urgence": "urge",
            "Ordonnance": "ordo",
            "CRH": "crh",
        },
        care_site_ids: List[int] = None,
    ):
        """Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

        Parameters
        ----------
        data : Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        care_site_relationship : pd.DataFrame
            DataFrame computed in the [``compute()``][edsteva.probes.base.BaseProbe.compute] that gives the hierarchy of the care site structure.
        extra_data : Data, optional
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]. This is not OMOP-standardized data but data needed to associate note with UF and Pole. If not provided, it will only compute the predictor for hospitals.
        start_date : datetime, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : datetime, optional
            **EXAMPLE**: `"2021-07-01"`
        care_site_levels : List[str], optional
            **EXAMPLE**: `["Hospital", "Pole", "UF"]`
        care_site_short_names : List[str], optional
            **EXAMPLE**: `["HOSPITAL 1", "HOSPITAL 2"]`
        stay_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Urg_and_consult": "urgences|consultation"}` or `"hospitalisés`
        note_types : Union[str, Dict[str, str]], optional
        care_site_ids : List[int], optional
            **EXAMPLE**: `[8312056386, 8312027648]`
        """
        check_tables(data=data, required_tables=["note"])

        visit_occurrence = prepare_visit_occurrence(
            data,
            start_date,
            end_date,
            stay_types,
        )

        care_site = prepare_care_site(
            data,
            care_site_ids,
            care_site_short_names,
            care_site_relationship,
        )

        note = prepare_note(data, note_types)

        hospital_visit = get_hospital_visit(note, visit_occurrence, care_site)
        hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
        note_predictor_by_level = {hospital_name: hospital_visit}

        # UF selection
        if not hospital_only(care_site_levels=care_site_levels) and extra_data:
            visit_detail = prepare_visit_detail(data, start_date, end_date)

            uf_visit = get_uf_visit(
                extra_data,
                note,
                visit_occurrence,
                visit_detail,
                care_site,
                care_site_relationship,
            )
            uf_name = CARE_SITE_LEVEL_NAMES["UF"]
            note_predictor_by_level[uf_name] = uf_visit

            pole_visit = get_pole_visit(uf_visit, care_site, care_site_relationship)
            pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
            note_predictor_by_level[pole_name] = pole_visit

        # Concatenate all predictors
        note_predictor = concatenate_predictor_by_level(
            predictor_by_level=note_predictor_by_level,
            care_site_levels=care_site_levels,
        )

        note_predictor = note_predictor.drop_duplicates(
            ["visit_id", "care_site_id", "stay_type", "note_type", "date"]
        )

        if is_koalas(note_predictor):
            note_predictor.spark.cache()

        return compute_completeness(note_predictor)



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd
from loguru import logger

from edsteva.probes.base import BaseProbe
from edsteva.probes.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    hospital_only,
    prepare_care_site,
    prepare_condition_occurrence,
    prepare_visit_occurrence,
)
from edsteva.utils.checks import check_tables
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data


def compute_completeness(condition_predictor):

    partition_cols = [
        "care_site_level",
        "care_site_id",
        "care_site_short_name",
        "stay_type",
        "diag_type",
        "condition_type",
        "date",
    ]
    n_visit_with_condition = (
        condition_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"has_condition": "count"})
        .rename(columns={"has_condition": "n_visit_with_condition"})
    )
    partition_cols = list(set(partition_cols) - {"diag_type", "condition_type"})

    n_visit = (
        condition_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"visit_id": "nunique"})
        .rename(columns={"visit_id": "n_visit"})
    )

    condition_predictor = n_visit_with_condition.merge(
        n_visit,
        on=partition_cols,
    )

    condition_predictor = to("pandas", condition_predictor)

    condition_predictor["c"] = condition_predictor["n_visit"].where(
        condition_predictor["n_visit"] == 0,
        condition_predictor["n_visit_with_condition"] / condition_predictor["n_visit"],
    )
    condition_predictor = condition_predictor.drop(columns=["n_visit_with_condition"])

    return condition_predictor


def get_hospital_visit(condition_occurrence, visit_occurrence, care_site, source):
    # visit/condition linkage
    if source == "AREM":
        # Link with visit_occurrence_source_value
        condition_hospital = condition_occurrence.drop_duplicates(
            ["visit_occurrence_source_value", "diag_type", "condition_type"]
        )
        condition_hospital["has_condition"] = True
        hospital_visit = condition_hospital.merge(
            visit_occurrence,
            on="visit_occurrence_source_value",
            how="left",
        ).drop(columns="visit_occurrence_source_value")
    else:
        condition_hospital = condition_occurrence.drop_duplicates(
            ["visit_occurrence_id", "diag_type", "condition_type"]
        )
        condition_hospital["has_condition"] = True
        hospital_visit = condition_hospital.merge(
            visit_occurrence,
            on="visit_occurrence_id",
            how="left",
        )
    hospital_visit = hospital_visit.rename(columns={"visit_occurrence_id": "visit_id"})
    hospital_visit = hospital_visit.merge(care_site, on="care_site_id")

    if is_koalas(hospital_visit):
        hospital_visit.spark.cache()

    return hospital_visit


class ConditionProbe(BaseProbe):
    r"""
    The [``ConditionProbe``][edsteva.probes.condition.ConditionProbe] computes $c_{condition}(t)$ the availability of claim data in patients' administrative stay:

    $$
    c_{condition}(t) = \frac{n_{with\,condition}(t)}{n_{visit}(t)}
    $$

    Where $n_{visit}(t)$ is the number of administrative stays, $n_{with\,condition}$ the number of stays having at least one claim code (e.g. ICD-10) recorded and $t$ is the month.

    Attributes
    ----------
    _index: List[str]
        Variable from which data is grouped

        **VALUE**: ``["care_site_level", "stay_type", "diag_type", "condition_type", "care_site_id"]``
    """

    _index = [
        "care_site_level",
        "stay_type",
        "diag_type",
        "condition_type",
        "care_site_id",
    ]

    def compute_process(
        self,
        data: Data,
        care_site_relationship: pd.DataFrame,
        extra_data: Data = None,
        start_date: datetime = None,
        end_date: datetime = None,
        care_site_levels: List[str] = None,
        stay_types: Union[str, Dict[str, str]] = None,
        diag_types: Union[str, Dict[str, str]] = None,
        condition_types: Union[str, Dict[str, str]] = {
            "All": ".*",
            "Cancer": "C",
        },
        care_site_ids: List[int] = None,
        care_site_short_names: List[str] = None,
        source: str = "ORBIS",
    ):
        """Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

        Parameters
        ----------
        data : Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        care_site_relationship : pd.DataFrame
            DataFrame computed in the [``compute()``][edsteva.probes.base.BaseProbe.compute] that gives the hierarchy of the care site structure.
        start_date : datetime, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : datetime, optional
            **EXAMPLE**: `"2021-07-01"`
        care_site_levels : List[str], optional
            **EXAMPLE**: `["Hospital", "Pole", "UF"]`
        stay_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Urg_and_consult": "urgences|consultation"}` or `"hospitalisés"`
        diag_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "DP\DR": "DP|DR"}` or `"DP"`
        condition_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Pulmonary_embolism": "I26"}`
        care_site_ids : List[int], optional
            **EXAMPLE**: `[8312056386, 8312027648]`
        care_site_short_names : List[str], optional
            **EXAMPLE**: `["HOSPITAL 1", "HOSPITAL 2"]`
        """

        check_tables(data=data, required_tables=["condition_occurrence"])

        if not hospital_only(care_site_levels=care_site_levels):
            logger.warning("Claim data is only available at hospital level")
            care_site_levels = None

        visit_occurrence = prepare_visit_occurrence(
            data=data,
            start_date=start_date,
            end_date=end_date,
            stay_types=stay_types,
        )

        condition_occurrence = prepare_condition_occurrence(
            data=data,
            extra_data=extra_data,
            source=source,
            diag_types=diag_types,
            condition_types=condition_types,
        )

        care_site = prepare_care_site(
            data=data,
            care_site_ids=care_site_ids,
            care_site_short_names=care_site_short_names,
            care_site_relationship=care_site_relationship,
        )

        hospital_visit = get_hospital_visit(
            condition_occurrence,
            visit_occurrence,
            care_site,
            source,
        )
        hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
        condition_predictor_by_level = {hospital_name: hospital_visit}

        condition_predictor = concatenate_predictor_by_level(
            predictor_by_level=condition_predictor_by_level,
            care_site_levels=care_site_levels,
        )

        return compute_completeness(condition_predictor)



import datetime
from abc import ABCMeta, abstractmethod
from typing import Dict, List, Union

import pandas as pd
from loguru import logger

from edsteva import CACHE_DIR
from edsteva.probes.utils import (
    delete_object,
    filter_table_by_care_site,
    get_care_site_relationship,
    load_object,
    save_object,
)
from edsteva.utils.checks import check_columns, check_tables
from edsteva.utils.typing import Data


class BaseProbe(metaclass=ABCMeta):
    """Base class for Probes

    Attributes
    ----------
    _schema: List[str]
        The columns a predictor must have

        **VALUE**: ``["care_site_id", "care_site_level", "stay_type", "date", "c"]``
    predictor: pd.DataFrame
        Available with the [``compute()``][edsteva.probes.base.BaseProbe.compute] method
    _cache_predictor: pd.DataFrame
        Available with the [``compute()``][edsteva.probes.base.BaseProbe.compute] method

        It is a copy of the predictor DataFrame used to [``reset_predictor()``][edsteva.probes.base.BaseProbe.reset_predictor]
    care_site_relationship: pd.DataFrame
        Available with the [``compute()``][edsteva.probes.base.BaseProbe.compute] method

        It describes the care site structure (cf. [``get_care_site_relationship()``][edsteva.probes.utils.get_care_site_relationship])
    """

    def __init__(self):
        self.is_valid_probe()
        self.name = self._get_name()

    _schema = ["care_site_id", "care_site_level", "stay_type", "date", "c"]

    def validate_input_data(self, data: Data) -> None:
        """Raises an error if the input data is not valid

        Parameters
        ----------
        data: Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        """

        if not isinstance(data, Data.__args__):
            raise TypeError("Unsupported type {} for data".format(type(data)))

        check_tables(
            data=data,
            required_tables=[
                "visit_occurrence",
                "care_site",
                "fact_relationship",
            ],
        )

    def is_valid_probe(self) -> None:
        """Raises an error if the instantiated Probe is not valid"""
        if not hasattr(self, "_index"):
            raise Exception(
                "Probe must have _index attribute. Please review the code of your probe"
            )

    def is_computed_probe(self) -> None:
        """Raises an error if the Probe has not been computed properly"""
        if hasattr(self, "predictor"):
            if not isinstance(self.predictor, pd.DataFrame):
                raise TypeError(
                    "Predictor must be a Pandas DataFrame and not a {}, please review the process method or your arguments".format(
                        type(self.predictor)
                    )
                )
            if self.predictor.empty:
                raise Exception(
                    "Predictor is empty, please review the process method or your arguments"
                )
            check_columns(
                self.predictor,
                required_columns=self._schema,
            )
            if not self.predictor.dtypes["date"] == "datetime64[ns]":
                try:
                    self.predictor["date"] = self.predictor["date"].astype(
                        "datetime64[ns]"
                    )
                except Exception as e:
                    raise TypeError(
                        "Predictor column 'date' type is {} and cannot convert to datetime and return the following error: {}. Please review the process method or your arguments".format(
                            self.predictor.dtypes["date"], e
                        )
                    )
        else:
            raise Exception(
                "Predictor has not been computed, please use the compute method as follow: Predictor.compute()"
            )

    def impute_missing_date(
        self,
        only_impute_per_care_site: bool = False,
    ) -> pd.DataFrame:
        """Impute missing date with 0 on the predictor of a probe.

        Parameters
        ----------
        only_impute_per_care_site : bool, optional
            If True it will only impute missing date between the first and the last observation of each care site.
            If False it will impute missing data on the entire study period whatever the care site
        """
        # Check if probe has been computed.
        self.is_computed_probe()

        # Set start_date to the beginning of the month.
        date_index = pd.date_range(
            start=self.start_date,
            end=self.end_date,
            freq="MS",
            closed="left",
        )
        date_index = pd.DataFrame({"date": date_index})

        # Precompute the mapping:
        # {'Hôpital-1': {'min': Timestamp('2010-06-01'), 'max': Timestamp('2019-11-01')}
        if only_impute_per_care_site:
            site_to_min_max_ds = (
                self.predictor.groupby(["care_site_short_name"])["date"]
                .agg([min, max])
                .to_dict("index")
            )

        partition_cols = self._index + ["care_site_short_name"]
        groups = []
        for partition, group in self.predictor.groupby(partition_cols):
            group = date_index.merge(group, on="date", how="left")

            # Filter on each care site timeframe.
            if only_impute_per_care_site:
                care_site_short_name = partition[-1]
                ds_min = site_to_min_max_ds[care_site_short_name]["min"]
                ds_max = site_to_min_max_ds[care_site_short_name]["max"]
                group = group.loc[(group["date"] >= ds_min) & (group["date"] <= ds_max)]

            # Fill specific partition values.
            for key, val in zip(partition_cols, partition):
                group[key] = val
            # Fill remaining NaN from counts values with 0.
            group.fillna(0, inplace=True)
            groups.append(group)

        self.predictor = pd.concat(groups)

    @abstractmethod
    def compute_process(
        self,
        data: Data,
        care_site_relationship: pd.DataFrame,
        start_date: datetime = None,
        end_date: datetime = None,
        care_site_levels: List[str] = None,
        stay_types: Union[str, Dict[str, str]] = None,
        care_site_ids: List[int] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Process the data in order to obtain a predictor table"""

    def compute(
        self,
        data: Data,
        start_date: datetime = None,
        end_date: datetime = None,
        care_site_levels: List[str] = None,
        stay_types: Union[str, Dict[str, str]] = None,
        care_site_ids: List[int] = None,
        impute_missing_dates: bool = True,
        only_impute_per_care_site: bool = False,
        **kwargs,
    ) -> None:
        """Calls [``compute_process()``][edsteva.probes.base.BaseProbe.compute_process]


        Here are the following computation steps:

        - check if input data is valid with [``validate_input_data()``][edsteva.probes.base.BaseProbe.validate_input_data] method
        - query care site relationship table with [``get_care_site_relationship()``][edsteva.probes.utils.get_care_site_relationship]
        - compute predictor with [``compute_process()``][edsteva.probes.base.BaseProbe.compute_process] method
        - check if predictor is valid with [``is_computed_probe()``][edsteva.probes.base.BaseProbe.is_computed_probe] method



        Parameters
        ----------
        data : Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        start_date : datetime, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : datetime, optional
            **EXAMPLE**: `"2021-07-01"`
        care_site_levels : List[str], optional
            **EXAMPLE**: `["Hospital", "Pole", "UF"]`
        stay_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Urg_and_consult": "urgences|consultation"}` or `"hospitalisés`
        care_site_ids : List[int], optional
            **EXAMPLE**: `[8312056386, 8312027648]`

        Attributes
        ----------
        Add to the Probe th following attributes:

        - predictor is the target DataFrame
        - _cache_predictor is a copy of the target DataFrame (used to [``reset_predictor()``][edsteva.probes.base.BaseProbe.reset_predictor])
        - care_site_relationship is a DataFrame with the hierarchy of the care site structure

        Examples
        -------

        ```python
        from edsteva.probes import VisitProbe

        visit = VisitProbe()
        visit.compute(
            data,
            stay_types={"All": ".*", "Urg_and_consult": "urgences|consultation"},
            care_site_levels=["Hospital", "Pole", "UF"],
        )
        visit.predictor.head()
        ```

        | care_site_level          | care_site_id | care_site_short_name | stay_type       | date       | n_visit | c     |
        | :----------------------- | :----------- | :------------------- | :-------------- | :--------- | :------ | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | Urg_and_consult | 2019-05-01 | 233.0   | 0.841 |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'All'           | 2021-04-01 | 393.0   | 0.640 |
        | Pôle/DMU                 | 8312027648   | Care site 2          | Urg_and_consult | 2011-03-01 | 204.0   | 0.497 |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'All'           | 2018-08-01 | 22.0    | 0.274 |
        | Hôpital                  | 8312022130   | Care site 3          | Urg_and_consult | 2022-02-01 | 9746.0  | 0.769 |


        """
        self.validate_input_data(data=data)
        care_site_relationship = get_care_site_relationship(data=data)

        self.predictor = self.compute_process(
            data=data,
            care_site_relationship=care_site_relationship,
            start_date=start_date,
            end_date=end_date,
            care_site_levels=care_site_levels,
            stay_types=stay_types,
            care_site_ids=care_site_ids,
            **kwargs,
        )
        self.is_computed_probe()

        self.start_date = (
            pd.to_datetime(start_date) if start_date else self.predictor["date"].min()
        )
        self.end_date = (
            pd.to_datetime(end_date) if end_date else self.predictor["date"].max()
        )

        if impute_missing_dates:
            self.impute_missing_date(
                only_impute_per_care_site=only_impute_per_care_site,
            )
        self.cache_predictor()
        self.care_site_relationship = care_site_relationship

    def reset_predictor(
        self,
    ) -> None:
        """Reset the predictor to its initial state"""
        self.predictor = self._cache_predictor.copy()

    def cache_predictor(
        self,
    ) -> None:
        """Cache the predictor"""
        self._cache_predictor = self.predictor.copy()
        logger.info(
            "Cache the predictor, you can reset the predictor to this state with the method reset_predictor"
        )

    def filter_care_site(
        self,
        care_site_ids: Union[int, List[int]] = None,
        care_site_short_names: Union[str, List[str]] = None,
    ) -> None:
        """Filters all the care sites related to the selected care sites.

        Parameters
        ----------
        care_site_ids : Union[int, List[int]], optional
            **EXAMPLE**: `[8312056386, 8312027648]`
        care_site_short_names : Union[str, List[str]], optional
            **EXAMPLE**: `["HOSPITAL 1", "HOSPITAL 2"]`
        """
        self.predictor = filter_table_by_care_site(
            table_to_filter=self.predictor,
            table_name="{} predictor".format(type(self).__name__.lower()),
            care_site_relationship=self.care_site_relationship,
            care_site_ids=care_site_ids,
            care_site_short_names=care_site_short_names,
        )
        logger.info("Use probe.reset_predictor() to get back the initial predictor")

    def load(self, path=None) -> None:
        """Loads a Probe from local

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.load(path=probe_path)
        ```

        """

        path = path or self._get_path()

        loaded_probe = load_object(path)
        self.__dict__ = loaded_probe.__dict__.copy()
        self.path = path

    def save(self, path: str = None, name: str = None) -> bool:
        """Saves computed Model instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        name : str, optional
            **EXAMPLE**: `"visit_from_BCT"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.compute(data)
        visit.save(path=probe_path)
        ```

        """

        self.is_computed_probe()

        if not path:
            if name:
                self.name = name
            path = self._get_path()

        self.path = path
        save_object(self, path)

    def delete(self, path: str = None):
        """Delete the saved Probe instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        """

        if not path:
            if hasattr(self, "path"):
                path = self.path
            else:
                path = self._get_path()

        delete_object(self, path)

    def _get_path(self):
        base_path = CACHE_DIR / "edsteva" / "probes"
        filename = f"{self.name.lower()}.pickle"
        return base_path / filename

    def _get_name(self):
        return type(self).__name__



import uuid

import altair as alt
from IPython.display import HTML, display

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.viz.dashboards.predictor_dashboard.fitted_probe import (
    fitted_probe_dashboard,
)
from edsteva.viz.dashboards.predictor_dashboard.probe import probe_dashboard
from edsteva.viz.utils import filter_predictor, save_html


def predictor_dashboard(
    probe: BaseProbe,
    fitted_model: BaseModel = None,
    care_site_level: str = None,
    save_path: str = None,
    x_axis_title: str = "Time (Month Year)",
    x_grid: bool = True,
    y_axis_title: str = "Completeness predictor c(t)",
    y_grid: bool = True,
    show_n_visit: bool = False,
    labelAngle: float = -90,
    labelFontSize: float = 12,
    titleFontSize: float = 13,
):
    r"""Displays an interactive chart with:

    - On the top, the aggregated average completeness predictor $c(t)$ over time $t$ with the fitted model $\hat{c}(t)$ if specified.
    - On the bottom, interactive filters including all the concepts in the [Probe][probe] (such as time, care site, number of visits...etc.)

    Is is possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$.
    fitted_model : BaseModel, optional
        Model fitted to the probe.
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    save_path : str, optional
        Folder path where to save the chart in HTML format.

        **EXAMPLE**: `"my_folder/my_file.html"`
    x_axis_title: str, optional,
        Label name for the x axis.
    x_grid: bool, optional,
        If False, remove the grid for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    y_grid: bool, optional,
        If False, remove the grid for the y axis.
    show_n_visit: bool, optional
        show the number of visit instead of the completeness predictor $c(t)$.
    labelAngle: float, optional
        The rotation angle of the label on the x_axis.
    labelFontSize: float, optional
        The font size of the labels (axis and legend).
    titleFontSize: float, optional
        The font size of the titles.
    """

    index = list(set(probe._index).difference(["care_site_level", "care_site_id"]))

    if fitted_model:
        predictor = fitted_model.predict(probe)
    else:
        predictor = probe.predictor

    predictor = filter_predictor(
        predictor=predictor,
        care_site_level=care_site_level,
    )

    if fitted_model:
        chart = fitted_probe_dashboard(
            predictor=predictor,
            index=index,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            x_grid=x_grid,
            y_grid=y_grid,
            labelAngle=labelAngle,
            labelFontSize=labelFontSize,
        )
    else:
        chart = probe_dashboard(
            predictor=predictor,
            index=index,
            x_axis_title=x_axis_title,
            y_axis_title=y_axis_title,
            x_grid=x_grid,
            y_grid=y_grid,
            show_n_visit=show_n_visit,
            labelAngle=labelAngle,
        )

    chart = chart.configure_axis(
        labelFontSize=labelFontSize, titleFontSize=titleFontSize
    ).configure_legend(labelFontSize=labelFontSize)
    vis_probe = "id" + uuid.uuid4().hex
    new_index_probe_id = "id" + uuid.uuid4().hex
    old_index_probe_id = "id" + uuid.uuid4().hex

    html_chart = f"""
        <!DOCTYPE html>
        <html>
        <head>
          <script src="https://cdn.jsdelivr.net/npm/vega@{alt.VEGA_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-lite@{alt.VEGALITE_VERSION}"></script>
          <script src="https://cdn.jsdelivr.net/npm/vega-embed@{alt.VEGAEMBED_VERSION}"></script>
        </head>
        <body>

        <div class="container">
          <div class="row">
            <div>
            <div id={vis_probe}></div>
            </div>
            <div style="position:absolute;left:45px;top:380px;width: -webkit-fill-available;">
            <div id={new_index_probe_id}>
              <div id={old_index_probe_id}></div>
            </div>
            <hr/>
            <h1 style="text-align:center"> Interactive filters </h1>
            </div>
          </div>
        </div>

        <script type="text/javascript">
        vegaEmbed('#{vis_probe}', {chart.to_json(indent=None)}).then(function(result) {{
            const sliders = document.getElementsByClassName('vega-bindings');
            const newparent = document.getElementById('{new_index_probe_id}');
            const oldchild = document.getElementById('{old_index_probe_id}');
            for (var i = 0; i < sliders.length; i++) {{
                if (sliders[i].parentElement.parentElement.id == '{vis_probe}') {{
                    var index_slider = sliders[i]
                    }}
                }}
            newparent.replaceChild(index_slider, oldchild);
            }}).catch(console.error);
        </script>
        </body>
        </html>
        """
    display(HTML(html_chart))
    if save_path:
        save_html(
            obj=html_chart,
            filename=save_path,
        )



from datetime import datetime
from typing import List, Union

import altair as alt
import pandas as pd

from edsteva.models.base import BaseModel
from edsteva.probes.base import BaseProbe
from edsteva.viz.utils import filter_predictor, round_it, save_html, scale_it


def plot_normalized_probe(
    probe: BaseProbe,
    fitted_model: BaseModel,
    t_0_max: datetime = None,
    t_1_min: datetime = None,
    error_max: float = None,
    c_0_min: float = None,
    care_site_level: str = None,
    stay_type: List[str] = None,
    care_site_id: List[int] = None,
    start_date: Union[datetime, str] = None,
    end_date: Union[datetime, str] = None,
    care_site_short_name: List[int] = None,
    t_min: int = None,
    t_max: int = None,
    save_path: str = None,
    x_axis_title: str = "Δt = (t - t₀) months",
    y_axis_title: str = "c(Δt) / c₀",
    show_per_care_site: bool = False,
    labelFontSize: float = 12,
    titleFontSize: float = 13,
    **kwargs,
):
    r"""Displays a chart with the aggregated normalized completeness predictor $\frac{c(\Delta t)}{c_0}$ over normalized time $\Delta t = t - t_0$. It represents the overall deviation from the Model.

    Is is possible to save the chart in HTML with the "save_path" optional input.

    Parameters
    ----------
    probe : BaseProbe
        Class describing the completeness predictor $c(t)$
    fitted_model : BaseModel
        Model fitted to the probe
    t_0_max : datetime, optional
        Initial value for the $t_0$ threshold. If None, it will be set as the maximum possible $t_0$ value.

        **EXAMPLE**: `"2022-01"`, `datetime(2020, 2, 1)`
    error_max : float, optional
        Initial value for the $error$ threshold. If None, it will be set as the maximum possible error value.

        **EXAMPLE**: `0.02`, `0.25`
    c_0_min : float, optional
        Initial value for the $c_0$ threshold. If None, it will be set as 0.

        **EXAMPLE**: `0.1`, `0.8`
    stay_type : List[str], optional
        **EXAMPLE**: `"All"` or `["All", "Urg"]`
    care_site_id : List[int], optional
        **EXAMPLE**: `[8312056386, 8312027648]`
    care_site_level : str, optional
        **EXAMPLE**: `"Hospital"`, `"Hôpital"` or `"UF"`
    stay_type : List[str], optional
        **EXAMPLE**: `"All"` or `["All", "Urg"]`
    care_site_id : List[int], optional
        **EXAMPLE**: `[8312056386, 8312027648]`
    start_date : datetime, optional
        **EXAMPLE**: `"2019-05-01"`
    end_date : datetime, optional
        **EXAMPLE**: `"2021-07-01"`
    care_site_short_name : List[int], optional
        **EXAMPLE**: `"HOSPITAL XXXX"`
    save_path : str, optional
        Folder path where to save the chart in HTML format.

        **EXAMPLE**: `"my_folder/my_file.html"`
    x_axis_title: str, optional,
        Label name for the x axis
    y_axis_title: str, optional,
        Label name for the y axis
    show_per_care_site: bool, optional
        If True, the average completeness predictor $c(t)$ is computed for each care site independently. If False, it is computed over all care sites.
    labelFontSize: float, optional
        The font size of the labels (axis and legend).
    titleFontSize: float, optional
        The font size of the titles.
    """

    predictor = probe.predictor.copy()
    estimates = fitted_model.estimates.copy()
    predictor = predictor.merge(estimates, on=probe._index)

    def month_diff(x, y):
        end = x.dt.to_period("M").view(dtype="int64")
        start = y.dt.to_period("M").view(dtype="int64")
        return end - start

    predictor["date"] = predictor["date"].astype("datetime64[ns]")
    predictor["t_0"] = predictor["t_0"].astype("datetime64[ns]")
    predictor["normalized_date"] = month_diff(predictor["date"], predictor["t_0"])
    predictor["t_0"] = predictor["t_0"].astype(str)
    predictor["normalized_date"] = predictor["normalized_date"].astype(int)
    if t_min:
        predictor = predictor[predictor.normalized_date >= t_min]
    if t_max:
        predictor = predictor[predictor.normalized_date <= t_max]
    predictor["normalized_c"] = predictor["c"].mask(
        (predictor["normalized_date"] >= 0) & (predictor["c_0"] == 0), 1
    )
    predictor["normalized_c"] = predictor["normalized_c"].mask(
        (predictor["normalized_date"] >= 0) & (predictor["c_0"] > 0),
        predictor["c"] / predictor["c_0"],
    )

    predictor["legend_error_band"] = "Standard deviation"
    predictor["legend_model"] = type(fitted_model).__name__

    predictor["model"] = predictor["c_0"].where(predictor["normalized_date"] < 0, 1)
    predictor["model"] = predictor["model"].where(predictor["normalized_date"] >= 0, 0)

    # RectangleModel
    if fitted_model.name == "RectangleFunction":
        predictor = predictor[predictor.date < predictor.t_1]

    index = list(set(probe._index).difference(["care_site_level", "care_site_id"]))
    if show_per_care_site:
        index = index + ["care_site_short_name"]

    predictor = filter_predictor(
        predictor=predictor,
        care_site_level=care_site_level,
        stay_type=stay_type,
        care_site_id=care_site_id,
        care_site_short_name=care_site_short_name,
        start_date=start_date,
        end_date=end_date,
        **kwargs,
    )
    index = [variable for variable in index if len(predictor[variable].unique()) >= 2]

    if c_0_min:
        c_0_min = round_it(float(c_0_min), 2)
        if c_0_min > round_it(predictor.c_0.max(), 2):
            c_0_min = round_it(predictor.c_0.max(), 2)
        elif c_0_min < round_it(predictor.c_0.min(), 2):
            c_0_min = round_it(predictor.c_0.min(), 2)
    else:
        c_0_min = round_it(predictor.c_0.min(), 2)

    if error_max:
        error_max = round_it(float(error_max), 2)
        if error_max > predictor.error.max():
            error_max = round_it(predictor.error.max(), 2)
        elif error_max < round_it(predictor.error.min(), 2):
            error_max = round_it(predictor.error.min(), 2)
    else:
        error_max = round_it(predictor.error.max(), 2)

    t_0_max = (
        str(pd.to_datetime(t_0_max)) if t_0_max else predictor.t_0.astype(str).max()
    )

    c_0_min_slider = alt.binding_range(
        min=0,
        max=round_it(predictor.c_0.max(), 2),
        step=1 / 100,
        name="c₀ min: ",
    )
    c_0_min_selection = alt.selection_single(
        name="c_0_min",
        fields=["c_0_min"],
        bind=c_0_min_slider,
        init={"c_0_min": c_0_min},
    )
    t_0_slider = alt.binding(
        input="t_0",
        name="t₀ max: ",
    )
    t_0_selection = alt.selection_single(
        name="t_0",
        fields=["t_0"],
        bind=t_0_slider,
        init={"t_0": t_0_max},
    )
    if fitted_model.name == "RectangleFunction":
        t_1_min = (
            str(pd.to_datetime(t_1_min)) if t_1_min else predictor.t_1.astype(str).min()
        )
        t_1_slider = alt.binding(
            input="t_1",
            name="t₁ min: ",
        )
        t_1_selection = alt.selection_single(
            name="t_1",
            fields=["t_1"],
            bind=t_1_slider,
            init={"t_1": t_1_min},
        )
    error_max_slider = alt.binding_range(
        min=round_it(predictor.error.min(), 2),
        max=round_it(predictor.error.max(), 2),
        step=scale_it(predictor.error.max()) / 100,
        name="error max: ",
    )
    error_max_selection = alt.selection_single(
        name="error_max",
        fields=["error_max"],
        bind=error_max_slider,
        init={"error_max": error_max},
    )

    alt.data_transformers.disable_max_rows()

    base_chart = (
        alt.Chart(predictor).encode(
            x=alt.X(
                "normalized_date:Q",
                title=x_axis_title,
                scale=alt.Scale(nice=False),
            ),
        )
    ).properties(width=900, height=300)

    transform_chart = (
        base_chart.transform_joinaggregate(
            mean_c_0="mean(c_0)",
            mean_error="mean(error)",
            groupby=["care_site_short_name"] + index,
        )
        .transform_filter(alt.datum.mean_c_0 >= c_0_min_selection.c_0_min)
        .transform_filter(alt.datum.mean_error <= error_max_selection.error_max)
        .transform_filter(alt.datum.t_0 <= t_0_selection.t_0)
    )
    # RectangleModel
    if fitted_model.name == "RectangleFunction":
        transform_chart = transform_chart.transform_filter(
            alt.datum.t_1 >= t_1_selection.t_1
        )

    mean_line = transform_chart.mark_line().encode(
        y=alt.Y(
            "mean(normalized_c):Q",
        )
    )

    error_line = transform_chart.mark_errorband(extent="stdev").encode(
        y=alt.Y(
            "normalized_c:Q",
            title=y_axis_title,
        ),
        stroke=alt.Stroke(
            "legend_error_band",
            title="Error band",
            legend=alt.Legend(symbolType="square", orient="top"),
        ),
    )

    model_line = base_chart.mark_line(
        color="black", interpolate="step-after", strokeDash=[5, 5]
    ).encode(
        y="model:Q",
        strokeWidth=alt.StrokeWidth(
            "legend_model",
            title="Model line",
            legend=alt.Legend(orient="top", symbolDash=[2, 2]),
        ),
    )

    chart = mean_line + error_line + model_line

    if len(index) >= 2:
        index_selection = alt.selection_single(
            fields=["index"],
            bind=alt.binding_radio(
                name="Plot average completeness per: ", options=index
            ),
            init={"index": index[0]},
        )
        chart = (
            chart.transform_fold(index, as_=["index", "value"])
            .encode(
                color=alt.Color(
                    "value:N",
                    sort={"field": "c_0", "op": "min", "order": "descending"},
                    title=None,
                ),
            )
            .transform_filter(index_selection)
            .add_selection(index_selection)
        )

    elif len(index) == 1:
        chart = chart.encode(
            color=alt.Color(
                "{}:N".format(index[0]),
                sort={"field": "c_0", "op": "min", "order": "descending"},
            ),
        )

    chart = (
        chart.add_selection(error_max_selection)
        .add_selection(c_0_min_selection)
        .add_selection(t_0_selection)
    )
    # RectangleModel
    if fitted_model.name == "RectangleFunction":
        chart = chart.add_selection(t_1_selection)

    if save_path:
        save_html(
            obj=chart.configure_axis(
                labelFontSize=labelFontSize, titleFontSize=titleFontSize
            ).configure_legend(labelFontSize=labelFontSize),
            filename=save_path,
        )

    return chart



from typing import List

import altair as alt
import pandas as pd


def probe_line(
    predictor: pd.DataFrame,
    index: List[str],
    x_axis_title: str = "Time (Month Year)",
    y_axis_title: str = "Completeness predictor c(t)",
    x_grid: bool = True,
    y_grid: bool = True,
    show_n_visit: bool = False,
    labelAngle: float = -90,
):
    """Script to be used by [``plot_probe()``][edsteva.viz.plots.plot_probe.wrapper]

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe
    index : List[str]
        Variable from which data is grouped
    x_axis_title: str, optional,
        Label name for the x axis.
    x_grid: bool, optional,
        If False, remove the grid for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    y_grid: bool, optional,
        If False, remove the grid for the y axis.
    show_n_visit: bool, optional
        show the number of visit instead of the completeness predictor $c(t)$
    labelAngle: float, optional
        The rotation angle of the label on the x_axis.
    """

    chart = (
        alt.Chart(predictor)
        .mark_line()
        .encode(
            x=alt.X(
                f'yearmonth({"date"}):T',
                title=x_axis_title,
                axis=alt.Axis(tickCount="month", labelAngle=labelAngle, grid=x_grid),
            ),
            y=alt.Y(
                "sum(n_visit):Q" if show_n_visit else "mean(c):Q",
                title=y_axis_title,
                axis=alt.Axis(grid=y_grid),
            ),
        )
    ).properties(width=800, height=300)

    if len(index) >= 2:
        index_selection = alt.selection_single(
            fields=["index"],
            bind=alt.binding_radio(
                name="Plot average completeness per: ", options=index
            ),
            init={"index": index[0]},
        )
        chart = (
            chart.transform_fold(index, as_=["index", "value"])
            .encode(
                color=alt.Color(
                    "value:N",
                    sort={"field": "c", "op": "mean", "order": "descending"},
                ),
            )
            .add_selection(index_selection)
            .transform_filter(index_selection)
        ).properties(width=800, height=300)
    elif len(index) == 1:
        chart = (
            chart.encode(
                color=alt.Color(
                    "{}:N".format(index[0]),
                    sort={"field": "c", "op": "mean", "order": "descending"},
                ),
            )
        ).properties(width=800, height=300)

    return chart



from types import ModuleType
from typing import Optional

import pandas as _pandas
from databricks import koalas as _koalas

from edsteva.utils.typing import DataObject

VALID_FRAMEWORKS = {
    "pandas": _pandas,
    "koalas": _koalas,
}


def get_framework(obj: DataObject) -> Optional[ModuleType]:
    for _, framework in VALID_FRAMEWORKS.items():
        if obj.__class__.__module__.startswith(framework.__name__):
            return framework
    # raise ValueError(f"Object from unknown framework: {obj}")
    return None


def is_pandas(obj: DataObject) -> bool:
    return get_framework(obj) == _pandas


def is_koalas(obj: DataObject) -> bool:
    return get_framework(obj) == _koalas


def to(framework: str, obj: DataObject) -> DataObject:
    if framework == "koalas" or framework is _koalas:
        return koalas(obj)
    elif framework == "pandas" or framework is _pandas:
        return pandas(obj)
    else:
        raise ValueError(f"Unknown framework: {framework}")


def pandas(obj: DataObject) -> DataObject:
    if get_framework(obj) is _pandas:
        return obj
    try:
        return obj.to_pandas()
    except AttributeError:
        pass
    raise ValueError("Could not convert object to pandas.")


def koalas(obj: DataObject) -> DataObject:
    if get_framework(obj) is _koalas:
        return obj
    try:
        return obj.to_koalas()
    except AttributeError:
        pass

    # will raise ValueError if impossible
    return _koalas.from_pandas(obj)



from datetime import datetime
from typing import Dict, List, Union

import pandas as pd

from edsteva.probes.base import BaseProbe
from edsteva.probes.utils import (
    CARE_SITE_LEVEL_NAMES,
    concatenate_predictor_by_level,
    convert_table_to_pole,
    convert_table_to_uf,
    hospital_only,
    prepare_care_site,
    prepare_visit_detail,
    prepare_visit_occurrence,
)
from edsteva.utils.framework import is_koalas, to
from edsteva.utils.typing import Data


def compute_completeness(visit_predictor):

    partition_cols = [
        "care_site_level",
        "care_site_id",
        "care_site_short_name",
        "stay_type",
        "date",
    ]
    n_visit = (
        visit_predictor.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )
        .agg({"visit_id": "nunique"})
        .rename(columns={"visit_id": "n_visit"})
    )

    n_visit = to("pandas", n_visit)

    partition_cols = list(set(partition_cols) - {"date"})
    q_99_visit = (
        n_visit.groupby(
            partition_cols,
            as_index=False,
            dropna=False,
        )[["n_visit"]]
        .quantile(q=0.99)
        .rename(columns={"n_visit": "q_99_visit"})
    )

    visit_predictor = n_visit.merge(
        q_99_visit,
        on=partition_cols,
    )

    visit_predictor["c"] = visit_predictor["q_99_visit"].where(
        visit_predictor["q_99_visit"] == 0,
        visit_predictor["n_visit"] / visit_predictor["q_99_visit"],
    )
    visit_predictor = visit_predictor.drop(columns="q_99_visit")

    return visit_predictor


def get_hospital_visit(visit_occurrence, care_site):
    hospital_visit = visit_occurrence.rename(
        columns={"visit_occurrence_id": "visit_id"}
    )
    hospital_visit = hospital_visit.merge(care_site, on="care_site_id")

    if is_koalas(hospital_visit):
        hospital_visit.spark.cache()

    return hospital_visit


def get_uf_visit(visit_occurrence, visit_detail, care_site, care_site_relationship):
    visit_detail = visit_detail.merge(
        visit_occurrence[["visit_occurrence_id", "stay_type"]],
        on="visit_occurrence_id",
    ).drop(columns="visit_occurrence_id")

    uf_visit = convert_table_to_uf(
        table=visit_detail,
        table_name="visit_detail",
        care_site_relationship=care_site_relationship,
    )
    uf_visit = uf_visit.merge(care_site, on="care_site_id")
    uf_visit = uf_visit[uf_visit["care_site_level"] == CARE_SITE_LEVEL_NAMES["UF"]]
    if is_koalas(uf_visit):
        uf_visit.spark.cache()

    return uf_visit


def get_pole_visit(uf_visit, care_site, care_site_relationship):
    pole_visit = convert_table_to_pole(
        table=uf_visit.drop(columns=["care_site_short_name", "care_site_level"]),
        table_name="uf_visit",
        care_site_relationship=care_site_relationship,
    )

    pole_visit = pole_visit.merge(care_site, on="care_site_id")
    pole_visit = pole_visit[
        pole_visit["care_site_level"] == CARE_SITE_LEVEL_NAMES["Pole"]
    ]
    if is_koalas(pole_visit):
        pole_visit.spark.cache()

    return pole_visit


class VisitProbe(BaseProbe):
    r"""
    The ``VisitProbe`` computes $c_(t)$ the availability of administrative data related to visits for each care site according to time:

    $$
    c(t) = \frac{n_{visit}(t)}{n_{99}}
    $$

    Where $n_{visit}(t)$ is the number of visits, $n_{99}$ is the $99^{th}$ percentile of visits and $t$ is the month.

    Attributes
    ----------
    _index: List[str]
        Variable from which data is grouped

        **VALUE**: ``["care_site_level", "stay_type", "care_site_id"]``
    """

    _index = ["care_site_level", "stay_type", "care_site_id"]

    def compute_process(
        self,
        data: Data,
        care_site_relationship: pd.DataFrame,
        start_date: datetime = None,
        end_date: datetime = None,
        care_site_levels: List[str] = None,
        stay_types: Union[str, Dict[str, str]] = None,
        care_site_ids: List[int] = None,
        care_site_short_names: List[str] = None,
    ):
        """Script to be used by [``compute()``][edsteva.probes.base.BaseProbe.compute]

        Parameters
        ----------
        data : Data
            Instantiated [``HiveData``][edsteva.io.hive.HiveData], [``PostgresData``][edsteva.io.postgres.PostgresData] or [``LocalData``][edsteva.io.files.LocalData]
        care_site_relationship : pd.DataFrame
            DataFrame computed in the [``compute()``][edsteva.probes.base.BaseProbe.compute] that gives the hierarchy of the care site structure.
        start_date : datetime, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : datetime, optional
            **EXAMPLE**: `"2021-07-01"`
        care_site_levels : List[str], optional
            **EXAMPLE**: `["Hospital", "Pole", "UF"]`
        stay_types : Union[str, Dict[str, str]], optional
            **EXAMPLE**: `{"All": ".*"}` or `{"All": ".*", "Urg_and_consult": "urgences|consultation"}` or `"hospitalisés`
        care_site_ids : List[int], optional
            **EXAMPLE**: `[8312056386, 8312027648]`
        care_site_short_names : List[str], optional
            **EXAMPLE**: `["HOSPITAL 1", "HOSPITAL 2"]`
        """

        visit_occurrence = prepare_visit_occurrence(
            data,
            start_date,
            end_date,
            stay_types,
        )

        care_site = prepare_care_site(
            data,
            care_site_ids,
            care_site_short_names,
            care_site_relationship,
        )

        hospital_visit = get_hospital_visit(
            visit_occurrence,
            care_site,
        )
        hospital_name = CARE_SITE_LEVEL_NAMES["Hospital"]
        visit_predictor_by_level = {hospital_name: hospital_visit}

        if not hospital_only(care_site_levels=care_site_levels):
            visit_detail = prepare_visit_detail(data, start_date, end_date)

            uf_name = CARE_SITE_LEVEL_NAMES["UF"]
            uf_visit = get_uf_visit(
                visit_occurrence,
                visit_detail,
                care_site,
                care_site_relationship,
            )
            visit_predictor_by_level[uf_name] = uf_visit

            pole_name = CARE_SITE_LEVEL_NAMES["Pole"]
            pole_visit = get_pole_visit(
                uf_visit,
                care_site,
                care_site_relationship,
            )
            visit_predictor_by_level[pole_name] = pole_visit

        visit_predictor = concatenate_predictor_by_level(
            predictor_by_level=visit_predictor_by_level,
            care_site_levels=care_site_levels,
        )

        return compute_completeness(visit_predictor)



from abc import ABCMeta, abstractmethod
from functools import reduce
from typing import Callable, List

import pandas as pd

from edsteva import CACHE_DIR
from edsteva.probes.base import BaseProbe
from edsteva.probes.utils import (
    delete_object,
    filter_table_by_date,
    load_object,
    save_object,
)


class BaseModel(metaclass=ABCMeta):

    """Base class for Models

    Attributes
    ----------
     _coefs: List[str]
        The list of the Model coefficients
    estimates: pd.DataFrame
        Available with the [``fit()``][edsteva.models.base.BaseModel.fit] method
    _metrics: List[str]
        Available with the [``fit()``][edsteva.models.base.BaseModel.fit] method

        The list of computed metrics if any
    params: List[str]
        Available with the [``fit()``][edsteva.models.base.BaseModel.fit] method

        Ths list of extra keyword parameters used.
    """

    def __init__(self):
        self.is_valid_model()
        self.name = type(self).__name__

    def is_valid_model(self) -> None:
        """Raises an error if the instantiated Model is not valid"""
        if not hasattr(self, "_coefs"):
            raise Exception(
                "Model must have _coefs attribute. Please review the code of your model"
            )

    def is_computed_estimates(self) -> None:
        """Raises an error if the Probe has not been fitted properly"""
        if hasattr(self, "estimates"):
            if isinstance(self.estimates, pd.DataFrame):
                if len(self.estimates) == 0:
                    raise Exception(
                        "Estimates are empty, please review the process method or your arguments"
                    )
            else:
                raise Exception(
                    "The fit process must return a Pandas Dataframe and not {}".format(
                        type(self.estimates)
                    )
                )

        else:
            raise Exception(
                "Model has not been fitted, please use the fit method as follow: Model.fit()"
            )

    @abstractmethod
    def fit_process(
        self,
        predictor: pd.DataFrame,
        index: List[str] = None,
        **kwargs,
    ):
        """Fit the Probe in order to obtain estimates"""

    @abstractmethod
    def predict_process(
        self,
        prediction: pd.DataFrame,
        **kwargs,
    ):
        """Compute the predicted Probe"""

    def fit(
        self,
        probe: BaseProbe,
        metric_functions: List[Callable] = None,
        start_date: str = None,
        end_date: str = None,
        **kwargs,
    ) -> None:
        """Fit the model to the probe instance

        Parameters
        ----------
        probe : BaseProbe
            Target variable to be fitted
        metric_functions : List[Callable], optional
            Metrics to apply on the fitted Probe. By default it will apply the default metric specified in the model.

            **EXAMPLE**: `[error, error_after_t0]`
        start_date : str, optional
            **EXAMPLE**: `"2019-05-01"`
        end_date : str, optional
            **EXAMPLE**: `"2021-07-01"`

        Examples
        --------
        ```python
        from edsteva.models.step_function import StepFunction

        step_function_model = StepFunction()
        step_function_model.fit(probe)
        step_function_model.estimates.head()
        ```

        | care_site_level          | care_site_id | stay_type    | t_0        | c_0   | error |
        | :----------------------- | :----------- | :----------- | :--------- | :---- | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | 'Urg_Hospit' | 2019-05-01 | 0.397 | 0.040 |
        | Unité Fonctionnelle (UF) | 8312056386   | 'All'        | 2011-04-01 | 0.583 | 0.028 |
        | Pôle/DMU                 | 8312027648   | 'Urg_Hospit' | 2021-03-01 | 0.677 | 0.022 |
        | Pôle/DMU                 | 8312027648   | 'All'        | 2018-08-01 | 0.764 | 0.014 |
        | Hôpital                  | 8312022130   | 'Urg_Hospit' | 2022-02-01 | 0.652 | 0.027 |
        """
        if isinstance(probe, BaseProbe):
            probe.is_computed_probe()
        else:
            raise TypeError("Unsupported type {} for probe.".format(type(probe)))

        predictor = filter_table_by_date(
            table=probe.predictor,
            table_name="predictor",
            start_date=start_date,
            end_date=end_date,
        )
        index = probe._index

        estimates = self.fit_process(
            predictor=predictor,
            index=index,
            **kwargs,
        )

        metrics = self._compute_metrics(
            predictor=predictor,
            estimates=estimates,
            index=index,
            metric_functions=metric_functions,
        )

        if metrics is not None:
            self._metrics = list(metrics.columns.difference(index))
            self.estimates = estimates.merge(metrics, on=index)

        else:
            self.estimates = estimates

        self.is_computed_estimates()
        self.params = kwargs

    def predict(
        self,
        probe: BaseProbe,
    ) -> pd.DataFrame:
        """Computes the predicted probe by using the estimates

        Parameters
        ----------
        probe : BaseProbe
            Target variable to be predicted

        Examples
        --------
        ```python
        from edsteva.models.step_function import StepFunction

        step_function_model.predict(visit).head()
        ```

        | care_site_level          | care_site_id | care_site_short_name | stay_type    | date       | n_visit | c     | c_fit |
        | :----------------------- | :----------- | :------------------- | :----------- | :--------- | :------ | :---- | :---- |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'Urg_Hospit' | 2019-05-01 | 233.0   | 0.841 | 0.758 |
        | Unité Fonctionnelle (UF) | 8312056386   | Care site 1          | 'All'        | 2021-04-01 | 393.0   | 0.640 | 0.758 |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'Urg_Hospit' | 2011-03-01 | 204.0   | 0.497 | 0     |
        | Pôle/DMU                 | 8312027648   | Care site 2          | 'All'        | 2018-08-01 | 22.0    | 0.784 | 0.874 |
        | Hôpital                  | 8312022130   | Care site 3          | 'Urg_Hospit' | 2022-02-01 | 9746.0  | 0.974 | 0.912 |

        """

        predictor = probe.predictor
        index = probe._index

        prediction = self.predict_process(predictor=predictor, index=index)
        return prediction

    def load(self, path=None) -> None:
        """Loads a Model from local

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.load(path=probe_path)
        ```

        """

        if not path:
            path = CACHE_DIR / "edsteva" / "models" / f"{self.name.lower()}.pickle"

        loaded_model = load_object(path)
        self.__dict__ = loaded_model.__dict__.copy()
        self.path = path

    def save(self, path: str = None, name: str = None) -> bool:
        """Saves computed Model instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        name : str, optional
            **EXAMPLE**: `"fitted_visit"`

        Examples
        -------
        ```python
        from edsteva.probes import VisitProbe

        probe_path = "my_path/visit.pkl"

        visit = VisitProbe()
        visit.compute(data)
        visit.save(path=probe_path)
        ```

        """

        self.is_computed_estimates()

        if not path:
            if name:
                self.name = name
            path = CACHE_DIR / "edsteva" / "models" / f"{self.name.lower()}.pickle"

        self.path = path
        save_object(self, path)

    def delete(self, path: str = None) -> bool:
        """Delete the saved Model instance

        Parameters
        ----------
        path : str, optional
            **EXAMPLE**: `"my_folder/my_file.html"`
        """

        if not path:
            if hasattr(self, "path"):
                path = self.path
            else:
                path = CACHE_DIR / "edsteva" / "models" / f"{self.name.lower()}.pickle"

        delete_object(self, path)

    def _compute_metrics(
        self,
        predictor: pd.DataFrame,
        estimates: pd.DataFrame,
        index: List[str],
        metric_functions: List[Callable] = None,
    ):
        if metric_functions:
            if callable(metric_functions):
                metrics = metric_functions(
                    predictor=predictor, estimates=estimates, index=index
                )
            elif isinstance(metric_functions, list):
                metrics = []
                for metric_function in metric_functions:
                    if callable(metric_function):
                        metrics.append(
                            metric_function(
                                predictor=predictor, estimates=estimates, index=index
                            )
                        )
                    else:
                        raise TypeError(
                            "{} is not callable. The metrics input must be a list of callable functions".format(
                                type(metric_function)
                            )
                        )
                metrics = reduce(
                    lambda left, right: pd.merge(left, right, on=index), metrics
                )
            else:
                raise TypeError(
                    "{} is not callable. The metrics input must be a callable function".format(
                        type(metrics)
                    )
                )

        elif hasattr(self, "default_metrics"):
            metrics = self.default_metrics(
                predictor=predictor, estimates=estimates, index=index
            )
        else:
            metrics = None

        return metrics

    def is_predictable_probe(
        self,
        predictor: pd.DataFrame,
        index: List[str],
    ) -> pd.DataFrame:
        """Raises an error if the model has not been fitted on the input predictor.

        Parameters
        ----------
        predictor : pd.DataFrame
            Target DataFrame to be predicted
        index : List[str]
            List of the columns given by Probe._index

        Returns
        -------
        pd.DataFrame
            Predictor along with the fitted estimates

        Raises
        ------
        Exception
            Some indexes have no associated estimates, the model must be fitted on an adequate probe
        """
        prediction = predictor.merge(
            self.estimates, on=index, how="left", validate="many_to_one", indicator=True
        )
        if (prediction["_merge"] == "both").all():
            prediction = prediction.drop(columns="_merge")
            return prediction

        else:
            raise Exception(
                "Some indexes have no associated estimates, the model must be fitted on an adequate probe"
            )



from functools import reduce

import altair as alt

from edsteva.models import BaseModel
from edsteva.viz.utils import save_html


def plot_estimates_densities(
    fitted_model: BaseModel,
    save_path: str = None,
    labelFontSize: float = 12,
    titleFontSize: float = 13,
    **kwargs,
):
    r"""Displays the density plot with the associated box plot of each estimate and metric computed in the input model. It can help you to set the thresholds.


    Parameters
    ----------
    fitted_model : BaseModel
        Model with estimates of interest
        **EXAMPLE**: StepFunction Model with $(\hat{t_0}, \hat{c_0})$
    save_path : str, optional
        Folder path where to save the chart in HTML format.

        **EXAMPLE**: `"my_folder/my_file.html"`
    labelFontSize: float, optional
        The font size of the labels (axis and legend).
    titleFontSize: float, optional
        The font size of the titles.
    """
    alt.data_transformers.disable_max_rows()

    estimates = fitted_model.estimates.copy()

    indexes = list(
        estimates.columns.difference(fitted_model._coefs + fitted_model._metrics)
    )
    indexes.remove("care_site_id")

    quantitative_estimates = []
    time_estimates = []

    for estimate in fitted_model._coefs + fitted_model._metrics:
        if estimates[estimate].dtype == float or estimates[estimate].dtype == int:
            max_value = estimates[estimate].max()
            min_value = estimates[estimate].min()
            estimates[estimate] = round(estimates[estimate], 3)

            estimate_density = (
                alt.vconcat(
                    (
                        (
                            alt.Chart(estimates)
                            .transform_density(
                                estimate,
                                as_=[estimate, "Density"],
                                groupby=indexes,
                                extent=[min_value, max_value],
                                **kwargs,
                            )
                            .mark_area()
                            .encode(
                                x=alt.X(
                                    "{}:Q".format(estimate),
                                    title=None,
                                ),
                                y="Density:Q",
                            )
                        )
                        + alt.Chart(estimates)
                        .mark_rule(color="red")
                        .encode(x="median({}):Q".format(estimate))
                    ).properties(width=800, height=300),
                    (
                        alt.Chart(estimates)
                        .mark_tick()
                        .encode(x=alt.X("{}:Q".format(estimate), axis=None))
                    ),
                    spacing=0,
                )
                & (
                    alt.Chart(estimates)
                    .mark_boxplot()
                    .encode(
                        x="{}:Q".format(estimate),
                    )
                )
            ).resolve_scale(x="shared")
            quantitative_estimates.append(estimate_density)

        else:
            estimates[estimate] = estimates[estimate]

            estimates[estimate] = estimates[estimate].astype("datetime64[s]")
            estimate_density = (
                (
                    alt.Chart(estimates)
                    .transform_timeunit(estimate="yearmonth({})".format(estimate))
                    .mark_bar(size=10)
                    .encode(
                        x=alt.X(
                            "{}:T".format(estimate),
                            axis=alt.Axis(
                                tickCount="month",
                                format="%Y, %b",
                                labelAngle=-90,
                            ),
                            title=estimate,
                        ),
                        y=alt.Y(
                            "count({}):Q".format(estimate),
                            axis=alt.Axis(tickMinStep=1),
                        ),
                    )
                )
                + alt.Chart(estimates)
                .mark_rule(color="red")
                .encode(x="median({}):T".format(estimate))
            ).properties(width=800, height=300)
            time_estimates.append(estimate_density)

    estimates_densities = time_estimates + quantitative_estimates
    estimates_densities = reduce(
        lambda estimate_density_1, estimate_density_2: estimate_density_1
        & estimate_density_2,
        estimates_densities,
    )

    for index in indexes:
        index_selection = alt.selection_single(
            fields=[index],
            bind=alt.binding_select(name=index, options=estimates[index].unique()),
            init={index: estimates[index].unique()[0]},
        )

        estimates_densities = estimates_densities.add_selection(index_selection)
        estimates_densities = estimates_densities.transform_filter(index_selection)

    if save_path:
        save_html(
            obj=estimates_densities.configure_axis(
                labelFontSize=labelFontSize, titleFontSize=titleFontSize
            ).configure_legend(labelFontSize=labelFontSize),
            filename=save_path,
        )

    return estimates_densities



from typing import List, Union

from edsteva.utils.typing import Data, DataFrame


class MissingColumnError(Exception):
    """Exception raised when a concept is missing"""

    def __init__(
        self,
        required_columns: Union[List, dict],
        df_name: str = "",
    ):

        if isinstance(required_columns, dict):
            to_display_per_column = [
                f"- {column} ({msg})" if msg is not None else f"{column}"
                for column, msg in required_columns.items()
            ]
        else:
            to_display_per_column = [f"- {column}" for column in required_columns]
        str_to_display = "\n".join(to_display_per_column)

        if df_name:
            df_name = f" {df_name} "
        message = (
            f"The{df_name}DataFrame is missing some columns, "
            "namely:\n"
            f"{str_to_display}"
        )

        super().__init__(message)


class MissingTableError(Exception):
    """Exception raised when a table is missing in the Data"""

    def __init__(
        self,
        required_tables: Union[List, dict],
    ):

        if isinstance(required_tables, dict):
            to_display_per_concept = [
                f"- {concept} ({msg})" if msg is not None else f"{concept}"
                for concept, msg in required_tables.items()
            ]
        else:
            to_display_per_concept = [f"- {concept}" for concept in required_tables]
        str_to_display = "\n".join(to_display_per_concept)

        message = f"Data is missing some tables, namely:\n {str_to_display}"

        super().__init__(message)


def check_columns(df: DataFrame, required_columns: List[str], df_name: str = ""):
    present_columns = set(df.columns)
    missing_columns = set(required_columns) - present_columns
    if missing_columns:
        raise MissingColumnError(missing_columns, df_name=df_name)


def check_tables(data: Data, required_tables: List[str]):
    present_tables = set(data.available_tables)
    missing_tables = set(required_tables) - present_tables
    if missing_tables:
        raise MissingTableError(missing_tables)



from edsteva.viz.plots.estimates_densities import plot_estimates_densities
from edsteva.viz.plots.normalized_probe import plot_normalized_probe
from edsteva.viz.plots.plot_probe.wrapper import plot_probe



from edsteva.viz.dashboards.estimates_dashboard import estimates_dashboard
from edsteva.viz.dashboards.predictor_dashboard.wrapper import predictor_dashboard

__all__ = [
    "predictor_dashboard",
    "estimates_dashboard",
]



from functools import reduce
from typing import List

import altair as alt
import pandas as pd


def fitted_probe_dashboard(
    predictor: pd.DataFrame,
    index: List[str],
    x_axis_title: str = "Time (Month Year)",
    y_axis_title: str = "Completeness predictor c(t)",
    x_grid: bool = True,
    y_grid: bool = True,
    labelAngle: float = -90,
    labelFontSize: float = 12,
):
    r"""Script to be used by [``predictor_dashboard()``][edsteva.viz.dashboards.predictor_dashboard.wrapper]

    Parameters
    ----------
    predictor : pd.DataFrame
        $c(t)$ computed in the Probe with its prediction $\hat{c}(t)$
    index : List[str]
        Variable from which data is grouped
    x_axis_title: str, optional,
        Label name for the x axis.
    x_grid: bool, optional,
        If False, remove the grid for the x axis.
    y_axis_title: str, optional,
        Label name for the y axis.
    y_grid: bool, optional,
        If False, remove the grid for the y axis.
    labelAngle: float, optional
        The rotation angle of the label on the x_axis.
    labelFontSize: float, optional
        The font size of the labels (axis and legend).
    """
    alt.data_transformers.disable_max_rows()

    predictor["legend_predictor"] = "Predictor c(t)"
    predictor["legend_model"] = "Model f(t)"

    time_selection = alt.selection_interval(encodings=["x"])
    time_line = (
        alt.Chart(predictor)
        .mark_line()
        .encode(
            x=alt.X(
                f'yearmonth({"date"}):T',
                title=x_axis_title,
                axis=alt.Axis(tickCount="month", labelAngle=labelAngle),
            ),
            y=alt.Y(
                "mean(c):Q",
                title="c(t)",
            ),
        )
        .add_selection(time_selection)
    ).properties(width=900, height=50)

    care_site_selection = alt.selection_multi(fields=["care_site_short_name"])
    care_site_color = alt.condition(
        care_site_selection,
        alt.Color(
            "care_site_short_name:N",
            legend=None,
            sort={"field": "c", "op": "mean", "order": "descending"},
        ),
        alt.value("lightgray"),
    )
    predictor_hist = (
        alt.Chart(predictor)
        .mark_bar()
        .encode(
            y=alt.Y(
                "care_site_short_name:N",
                title="Care site short name",
                sort="-x",
            ),
            color=care_site_color,
        )
        .transform_filter(time_selection)
        .add_selection(care_site_selection)
    ).properties(width=300)

    care_site_hist = predictor_hist.encode(
        x=alt.X(
            "mean(c):Q",
            title="c",
        ),
        tooltip=alt.Tooltip("mean(c):Q", format=".2"),
    ).properties(title="Average completeness per care site")

    index_variables_hists = []
    index_variables_selections = []
    for index_variable in index:
        index_variable_selection = alt.selection_multi(fields=[index_variable])
        index_variables_selections.append(index_variable_selection)

        index_variable_color = alt.condition(
            index_variable_selection,
            alt.Color(
                "{}:N".format(index_variable),
                legend=None,
                sort={"field": "c", "op": "mean", "order": "descending"},
            ),
            alt.value("lightgray"),
        )
        index_variable_hist = (
            alt.Chart(predictor)
            .mark_bar()
            .encode(
                y=alt.Y(
                    "mean(c):Q",
                    title="c",
                ),
                x=alt.X(
                    "{}:N".format(index_variable),
                    title=index_variable,
                    sort="-y",
                ),
                tooltip=alt.Tooltip("mean(c):Q", format=".2"),
                color=index_variable_color,
            )
            .add_selection(index_variable_selection)
            .transform_filter(time_selection)
            .transform_filter(care_site_selection)
        ).properties(title="Average completeness per {}".format(index_variable))
        index_variables_hists.append(index_variable_hist)

    extra_predictors = predictor.columns.difference(
        index
        + [
            "date",
            "care_site_short_name",
            "care_site_id",
            "c",
            "c_hat",
            "legend_predictor",
            "legend_model",
        ]
    )
    extra_predictors_hists = []

    for extra_predictor in extra_predictors:

        extra_predictor_hist = predictor_hist.encode(
            x=alt.X(
                "sum({}):Q".format(extra_predictor),
                title="{}".format(extra_predictor),
                axis=alt.Axis(format="s"),
            ),
            tooltip=alt.Tooltip("sum({}):Q".format(extra_predictor), format=","),
        ).properties(title="Total {} per care site".format(extra_predictor))

        extra_predictors_hists.append(extra_predictor_hist)

    index.append("care_site_short_name")
    index_selection = alt.selection_single(
        fields=["index"],
        bind=alt.binding_radio(name="Plot average completeness per: ", options=index),
        init={"index": "stay_type"},
    )

    base_line = (
        alt.Chart(predictor)
        .transform_fold(index, as_=["index", "value"])
        .encode(
            x=alt.X(
                f'yearmonth({"date"}):T',
                title=x_axis_title,
                axis=alt.Axis(tickCount="month", labelAngle=-90, grid=x_grid),
            ),
            color=alt.Color(
                "value:N",
                sort={"field": "c", "op": "mean", "order": "descending"},
                title=None,
            ),
        )
        .transform_filter(index_selection)
    ).properties(width=900, height=300)

    probe_line = base_line.mark_line().encode(
        y=alt.Y(
            "mean(c):Q",
            title=y_axis_title,
            axis=alt.Axis(grid=y_grid),
        ),
        strokeDash=alt.StrokeDash(
            "legend_predictor",
            title="",
            legend=alt.Legend(
                symbolType="stroke",
                symbolStrokeColor="steelblue",
                labelFontSize=labelFontSize,
                labelFontStyle="bold",
                orient="left",
            ),
        ),
    )

    model_line = base_line.mark_line(
        interpolate="step-after", strokeDash=[5, 5]
    ).encode(
        y=alt.Y(
            "mean(c_hat):Q",
        ),
        strokeWidth=alt.StrokeWidth(
            "legend_model",
            title="",
            legend=alt.Legend(
                symbolType="stroke",
                symbolStrokeColor="steelblue",
                labelFontSize=labelFontSize,
                labelFontStyle="bold",
                symbolDash=[2, 2],
                orient="left",
            ),
        ),
    )

    fitted_probe = (probe_line + model_line).add_selection(index_selection)

    fitted_probe = fitted_probe.transform_filter(care_site_selection)
    fitted_probe = fitted_probe.transform_filter(time_selection)

    for index_variable_selection in index_variables_selections:
        fitted_probe = fitted_probe.transform_filter(index_variable_selection)
        care_site_hist = care_site_hist.transform_filter(index_variable_selection)
        for extra_predictor_hist in extra_predictors_hists:
            extra_predictor_hist = extra_predictor_hist.transform_filter(
                index_variable_selection
            )
        for idx in range(len(index_variables_hists)):
            if idx != index_variables_selections.index(index_variable_selection):
                index_variables_hists[idx] = index_variables_hists[
                    idx
                ].transform_filter(index_variable_selection)

    index_variables_hists = reduce(
        lambda index_variable_hist_1, index_variable_hist_2: index_variable_hist_1
        & index_variable_hist_2,
        index_variables_hists,
    )
    extra_predictors_hists = reduce(
        lambda extra_predictor_hist_1, extra_predictor_hist_2: extra_predictor_hist_1
        & extra_predictor_hist_2,
        extra_predictors_hists,
    )

    chart = (
        alt.vconcat(
            alt.vconcat(fitted_probe, time_line, spacing=130),
            (index_variables_hists | care_site_hist | extra_predictors_hists),
            spacing=10,
        )
    ).resolve_scale(color="independent")

    return chart



def recursive_items(dictionary):
    for key, value in dictionary.items():
        if type(value) is dict:
            if key[0] != "P":
                yield key
            yield from recursive_items(value)
        else:
            yield key



import os
from datetime import datetime
from math import ceil, floor, log10
from pathlib import Path
from typing import List, Union

import altair as alt
import pandas as pd
from loguru import logger

from edsteva.probes.utils import CARE_SITE_LEVEL_NAMES, filter_table_by_date


def save_html(obj: alt.Chart, filename: str):
    """Save chart in the specified file

    Parameters
    ----------
    obj : alt.Chart
        Altair chart to be saved
    filename : str
        Folder path where to save the chart in HTML format.

        **EXAMPLE**: `"my_folder/my_file.html"`
    """
    if not isinstance(filename, Path):
        filename = Path(filename)
    os.makedirs(filename.parent, exist_ok=True)
    if hasattr(obj, "save"):
        obj.save(filename)
    else:
        with open(filename, "w") as f:
            f.write(obj)
    logger.info("The chart has been saved in {}", filename)


def round_it(x: float, sig: int):
    if x == 0:
        return 0
    else:
        return round(x, sig - int(floor(log10(abs(x)))) - 1)


def scale_it(x: float):
    if x == 0:
        return 1
    else:
        return 10 ** ceil(log10(x))


def filter_predictor(
    predictor: pd.DataFrame,
    care_site_level: str = None,
    stay_type: List[str] = None,
    care_site_id: List[int] = None,
    care_site_short_name: List[int] = None,
    start_date: Union[datetime, str] = None,
    end_date: Union[datetime, str] = None,
    **kwargs
):

    # Time
    predictor = filter_table_by_date(
        table=predictor,
        table_name="predictor",
        start_date=start_date,
        end_date=end_date,
    )

    # Care site level
    if not care_site_level:
        if len(predictor.care_site_level.unique()) == 1:
            care_site_level = predictor.care_site_level.unique()[0]
        else:
            care_site_level = CARE_SITE_LEVEL_NAMES["Hospital"]
    elif care_site_level in CARE_SITE_LEVEL_NAMES.keys():
        care_site_level = CARE_SITE_LEVEL_NAMES[care_site_level]

    if care_site_level not in predictor.care_site_level.unique():
        raise AttributeError(
            "The selected care site level {} is not part of the computed care site levels {}".format(
                care_site_level, list(predictor.care_site_level.unique())
            )
        )
    predictor = predictor[predictor["care_site_level"] == care_site_level].drop(
        columns=["care_site_level"]
    )
    logger.debug(
        "Predictor has been filtered on the selected care site level : {}",
        care_site_level,
    )

    # Stay type
    if stay_type:
        if not isinstance(stay_type, list):
            stay_type = [stay_type]
        predictor = predictor[predictor.stay_type.isin(stay_type)]
        logger.debug(
            "Predictor has been filtered on the selected stay type : {}",
            stay_type,
        )

    # Care site id
    if care_site_id:
        if not isinstance(care_site_id, list):
            care_site_id = [care_site_id]
        predictor = predictor[predictor.care_site_id.isin(care_site_id)]
        logger.debug(
            "Predictor has been filtered on the selected care site id : {}",
            care_site_id,
        )

    # Care site short name
    if care_site_short_name:
        if not isinstance(care_site_short_name, list):
            care_site_short_name = [care_site_short_name]
        predictor = predictor[predictor.care_site_short_name.isin(care_site_short_name)]
        logger.debug(
            "Predictor has been filtered on the selected care site short name : {}",
            care_site_short_name,
        )

    # Others
    for key, value in kwargs.items():
        if not isinstance(value, list):
            value = [value]
        predictor = predictor[predictor[key].isin(value)]
        logger.debug(
            "Predictor has been filtered on the selected {} : {}",
            key,
            care_site_short_name,
        )
    return predictor
